import abc
import surfex
from surfex.util import error, warning
import copy
import numpy as np
from datetime import timedelta


class Variable(object):
    __metaclass__ = abc.ABCMeta

    """
    Variable top class
    The variable read it self
    """

    def __init__(self, basetime, validtime, var_dict, debug):
        intervall = 3600
        if "timstep" in var_dict:
            intervall = var_dict["timestep"]
        self.initialtime = validtime
        if validtime is not None:
            self.previoustime = validtime-timedelta(seconds=intervall)
        else:
            self.previoustime = None
        self.basetime = basetime
        self.validtime = validtime
        self.var_dict = copy.deepcopy(var_dict)
        if "filepattern" in var_dict:
            self.filepattern = var_dict["filepattern"]
        else:
            self.filepattern = None
        self.file_handler = None
        self.previousfilename = None
        if validtime is not None:
            self.time_elapsed = validtime-basetime
        else:
            self.time_elapsed = 0
        self.filename = surfex.file.parse_filepattern(self.filepattern, self.basetime, self.validtime)
        self.debug = debug
        if self.debug:
            print("Constructed " + self.__class__.__name__ + " for " + str(self.var_dict))

    @abc.abstractmethod
    def read_variable(self, geo, validtime, cache):
        raise NotImplementedError('users must define read_variable to use this base class')

    @abc.abstractmethod
    def print_variable_info(self):
        raise NotImplementedError('users must define print_variable_info to use this base class')

    @staticmethod
    def deaccumulate(field, previous_field, instant):

        field = np.subtract(field, previous_field)
        if any(field[field < 0.]):
            neg = []
            for i in range(0, field.shape[0]):
                if field[i] < 0.:
                    neg.append(field[i])
            neg = np.asarray(neg)
            warning("Deaccumulated field has " + str(neg.shape[0]) + " negative lowest:"
                    + str(np.nanmin(neg)) + " mean: " + str(np.nanmean(neg)))
        field[field < 0.] = 0
        if float(instant) != 0.:
            field = np.divide(field, float(instant))
        return field

    def open_new_file(self, fcint, offset, file_inc):

        if self.filepattern is None:
            return True

        if self.validtime is None:
            return True

        new = False
        filepattern = self.var_dict["filepattern"]
        basetime = self.basetime
        validtime = self.validtime
        new_basetime = basetime

        # Basetime checks
        if offset >= 0:
            # Change basetime if offset is exceeded
            if (validtime-basetime) > (timedelta(seconds=fcint)+timedelta(seconds=offset)):
                new_basetime = basetime+timedelta(seconds=fcint)
                if self.debug:
                    print("Changing basetime to ", new_basetime)
        else:
            raise Exception("Negative offset does not make sense here")

        # Always open the file for the first step
        if self.validtime == self.initialtime:
            if self.debug:
                print("Same as initial time ", self.initialtime)
            new = True

        # File increment checks
        if file_inc > 0:
            if file_inc > offset:
                if offset == 0:
                    if self.time_elapsed == timedelta(seconds=file_inc):
                        if self.debug:
                            print("Test for file_inc: ", self.time_elapsed,
                                  timedelta(seconds=file_inc) + timedelta(seconds=offset))
                        self.time_elapsed = timedelta(seconds=0)
                        new = True
                else:
                    if self.time_elapsed > (timedelta(seconds=file_inc) + timedelta(seconds=offset)):
                        if self.debug:
                            print("Test for file_inc: ", self.time_elapsed,
                                  timedelta(seconds=file_inc) + timedelta(seconds=offset))
                        self.time_elapsed = timedelta(seconds=0)
                        new = True

            else:
                if self.time_elapsed >= timedelta(seconds=file_inc):
                    if self.debug:
                        print("Test for file_inc: ", self.time_elapsed, timedelta(seconds=file_inc))
                    self.time_elapsed = timedelta(seconds=0)
                    new = True
        else:
            error("file_inc must be a positive value > 0")

        # Set filename. New basetime is the same as previous or the updated one
        self.filename = surfex.file.parse_filepattern(filepattern, new_basetime, validtime)
        self.previousfilename = surfex.file.parse_filepattern(filepattern, basetime, self.previoustime)
        # Reread if the basetime has changed
        if new_basetime != basetime:
            new = True
            self.previousfilename = self.filename

        self.time_elapsed = self.time_elapsed + (validtime - self.previoustime)
        self.basetime = new_basetime
        if new and self.debug:
            print("Open new file ", self.filename, self.validtime, self.basetime)
            print("Previous file is now", self.previousfilename, self.previoustime)

        return new

    def rotate_geographic_wind(self, field, interpolator):
        rotate_wind = False
        if "rotate_to_geographic" in self.var_dict:
            rotate_wind = self.var_dict["rotate_to_geographic"]
        if rotate_wind:
            field = interpolator.rotate_wind_to_geographic(field)
            return field
        else:
            return field


class NetcdfVariable(Variable):

    """
    NetCDF variable
    """

    def __init__(self, var_dict, basetime, validtime, debug):

        mandatory = ["name", "fcint", "offset", "file_inc", "filepattern"]
        for i in range(0, len(mandatory)):
            if mandatory[i] not in var_dict:
                raise Exception("NetCDF variable must have attribute " + mandatory[i] + " var_dict:" + str(var_dict))

        Variable.__init__(self, basetime, validtime, var_dict, debug)

    def read_variable(self, geo, validtime, cache):

        self.validtime = validtime
        if self.open_new_file(int(self.var_dict["fcint"]), int(self.var_dict["offset"]),
                              int(self.var_dict["file_inc"])):
            # print "Updating filehandler for "+self.print_variable_info()
            if cache is not None and cache.file_open(self.filename):
                self.file_handler = cache.get_file_handler(self.filename)
            else:
                self.file_handler = surfex.netcdf.Netcdf(self.filename)
                if cache is not None:
                    cache.set_file_handler(self.filename, self.file_handler)

        if self.file_handler is None:
            surfex.util.warning("No file handler exist for this time step")
            field = np.array([len(geo.lons)])
            field = field.fill(np.nan)
        else:
            var_name = self.var_dict["name"]
            level = None
            accumulated = False
            units = None
            if "level" in self.var_dict:
                level = [self.var_dict["level"]]
            if "units" in self.var_dict:
                units = str([self.var_dict["units"]][0])
            if "accumulated" in self.var_dict:
                accumulated = self.var_dict["accumulated"]
            int_type = "nearest"
            if "interpolator" in self.var_dict:
                int_type = self.var_dict["interpolator"]

            # Re-read field
            previous_field = None
            if accumulated:
                print(self.basetime, self.initialtime, self.previoustime)
                can_read = True
                if self.basetime <= self.initialtime:
                    if self.previoustime < self.basetime:
                        can_read = False

                if not can_read:
                    surfex.warning("Can not read previous time step for this time. Setting it to 0.")
                    print(self.basetime, self.initialtime, self.previoustime)
                    previous_field = np.zeros([geo.npoints])
                else:
                    # Re-read field
                    id_str = None
                    if cache is not None:
                        id_str = cache.generate_netcdf_id(var_name, self.previousfilename, self.previoustime)
                    if cache is not None and cache.is_saved(id_str):
                        print("Updating cached value ", id_str)
                        previous_field = cache.saved_fields[id_str]
                    else:
                        # Modify filename in handler
                        fname = self.filename
                        if self.debug:
                            print("Re-read ", self.previoustime, " from ", self.previousfilename)
                        self.file_handler.fname = self.previousfilename
                        previous_field, intp = self.file_handler.points(var_name,  geo, level=level,
                                                                        validtime=self.previoustime,
                                                                        interpolation=int_type,
                                                                        units=units, cache=cache)
                        if cache is not None:
                            cache.save_field(id_str, previous_field)
                        # Change filename back in handler. Ready to read this time step
                        self.file_handler.fname = fname

            field, interpolator = self.file_handler.points(var_name, geo, level=level, validtime=validtime,
                                                           interpolation=int_type, units=units, cache=cache)
            # Rotate wind to geographic if requested
            field = self.rotate_geographic_wind(field, interpolator)
            if cache is not None:
                id_str = cache.generate_netcdf_id(var_name, self.filename, validtime)
                print(id_str)
                cache.save_field(id_str, field)

            if accumulated:
                instant = [(validtime - self.previoustime).total_seconds()]
                if "instant" in self.var_dict:
                    instant = [self.var_dict["instant"]]
                field = self.deaccumulate(field, previous_field, float(instant[0]))

        self.previoustime = validtime
        return field

    def print_variable_info(self):
        print(":" + str(self.var_dict) + ":")


class GribVariable(Variable):

    """
    Grib variable
    """
    def __init__(self, var_dict, basetime, validtime, debug, grib_type="grib2"):
        if grib_type == "grib1":
            mandatory = ["parameter", "type", "level", "tri", "fcint", "offset", "file_inc", "filepattern"]
            print(var_dict)
            for i in range(0, len(mandatory)):
                if mandatory[i] not in var_dict:
                    error("Grib1 variable must have attribute \"" + mandatory[i] + "\". var_dict=" + str(var_dict))
            Variable.__init__(self, basetime, validtime, var_dict, debug)
            self.grib_type = grib_type
        elif grib_type == "grib2":
            mandatory = ["discipline", "parameterCategory", "parameterNumber", "levelType", "level",
                         "typeOfStatisticalProcessing", "fcint", "offset", "file_inc", "filepattern"]
            print(var_dict)
            for i in range(0, len(mandatory)):
                if mandatory[i] not in var_dict:
                    error("Grib2 variable must have attribute \"" + mandatory[i] + "\". var_dict=" + str(var_dict))
            Variable.__init__(self, basetime, validtime, var_dict, debug)
            self.grib_type = grib_type
        else:
            raise NotImplementedError

    def read_variable(self, geo, validtime, cache):
        self.validtime = validtime
        if self.open_new_file(int(self.var_dict["fcint"]), int(self.var_dict["offset"]),
                              int(self.var_dict["file_inc"])):

            # print "Updating filehandler for "+self.print_variable_info()
            if cache is not None and cache.file_open(self.filename):
                self.file_handler = cache.get_file_handler(self.filename)
            else:
                self.file_handler = surfex.grib.Grib(self.filename, debug=self.debug)
                if cache is not None:
                    cache.set_file_handler(self.filename, self.file_handler)

        if self.file_handler is None:
            warning("No file handler exist for this time step")
            field = np.array([len(geo.lons)])
            field = field.fill(np.nan)
        else:
            if self.grib_type == "grib1":
                par = self.var_dict["parameter"]
                typ = self.var_dict["type"]
                level = self.var_dict["level"]
                tri = self.var_dict["tri"]
                gribvar = surfex.grib.Grib1Variable(par, typ, level, tri, debug=self.debug)
            elif self.grib_type == "grib2":
                discipline = self.var_dict["discipline"]
                pc = self.var_dict["parameterCategory"]
                pn = self.var_dict["parameterNumber"]
                lt = self.var_dict["levelType"]
                lev = self.var_dict["level"]
                tsp = -1
                if "typeOfStatisticalProcessing" in self.var_dict:
                    tsp = self.var_dict["typeOfStatisticalProcessing"]
                gribvar = surfex.grib.Grib2Variable(discipline, pc, pn, lt, lev, tsp=tsp, debug=self.debug)
            else:
                raise NotImplementedError

            int_type = "nearest"
            if "interpolator" in self.var_dict:
                int_type = self.var_dict["interpolator"]

            # Re-read field
            previous_field = None
            if gribvar.is_accumulated():
                print(self.basetime, self.initialtime, self.previoustime)
                can_read = True
                if self.basetime <= self.initialtime:
                    if self.previoustime < self.basetime:
                        can_read = False

                if not can_read:
                    surfex.warning("Can not read previous time step for this time. Setting it to 0.")
                    print(self.basetime, self.initialtime, self.previoustime)
                    previous_field = np.zeros([geo.npoints])
                else:
                    id_str = None
                    if cache is not None:
                        id_str = cache.generate_grib_id(gribvar, self.previousfilename, self.previoustime)
                    if cache is not None and cache.is_saved(id_str):
                        previous_field = cache.saved_fields[id_str]
                    else:
                        # Modify filename in handler
                        fname = self.filename
                        if self.debug:
                            print("Re-read ", self.previoustime, " from ", self.previousfilename)
                        self.file_handler.fname = self.previousfilename
                        previous_field, intp = self.file_handler.points(gribvar, geo, self.previoustime,
                                                                        interpolation=int_type, cache=cache)

                        # Change filename back in handler. Ready to read this time step
                        self.file_handler.fname = fname
                        if cache is not None:
                            cache.save_field(id_str, previous_field)

            # Read field
            id_str = None
            if cache is not None:
                id_str = cache.generate_grib_id(gribvar, self.filename, self.validtime)
            if cache is not None and cache.is_saved(id_str):
                field = cache.saved_fields[id_str]
            else:
                field, interpolator = self.file_handler.points(gribvar, geo, validtime, interpolation=int_type,
                                                               cache=cache)
                # Rotate wind to geographic if requested
                field = self.rotate_geographic_wind(field, interpolator)
                if cache is not None:
                    cache.save_field(id_str, field)

            # Deaccumulate
            if gribvar.is_accumulated():
                instant = [(validtime - self.previoustime).total_seconds()]
                if "instant" in self.var_dict:
                    instant = [self.var_dict["instant"]]
                field = self.deaccumulate(field, previous_field, float(instant[0]))

        self.previoustime = validtime
        return field

    def print_variable_info(self):
        print(":" + str(self.var_dict) + ":")


class SurfexVariable(Variable):

    def __init__(self, var_dict, basetime, validtime, debug):
        mandatory = ["varname", "fcint", "offset", "file_inc", "filepattern"]
        # , "patches", "layers", "accumulated", "fcint", "offset", "file_inc", "filepattern",
        #         "fileformat", "filetype"]
        for i in range(0, len(mandatory)):
            print(mandatory[i], var_dict)
            if mandatory[i] not in var_dict:
                raise Exception("Surfex variable must have attribute " + mandatory[i] + " var_dict:" + str(var_dict))

        Variable.__init__(self, basetime, validtime, var_dict, debug)

    def read_variable(self, geo, validtime, cache):

        self.validtime = validtime
        if self.open_new_file(int(self.var_dict["fcint"]), int(self.var_dict["offset"]),
                              int(self.var_dict["file_inc"])):

            geo_in = None
            if "geo_input" in self.var_dict:
                geo_in = self.var_dict["geo_input"]

            # print "Updating filehandler for "+self.print_variable_info()
            if cache is not None and cache.file_open(self.filename):
                self.file_handler = cache.get_file_handler(self.filename)
            else:
                fileformat = None
                if "fileformat" in self.var_dict:
                    fileformat = self.var_dict["fileformat"]
                filetype = None
                if "filetype" in self.var_dict:
                    filetype = self.var_dict["filetype"]
                self.file_handler = surfex.file.get_surfex_io_object(self.filename, fileformat=fileformat,
                                                                     filetype=filetype, geo=geo_in)
                if cache is not None:
                    cache.set_file_handler(self.filename, self.file_handler)

        if self.file_handler is None:
            warning("No file handler exist for this time step")
            field = np.array([len(geo.lons)])
            field = field.fill(np.nan)
        else:
            varname = self.var_dict["varname"]
            layers = None
            if "layers" in self.var_dict:
                layers = self.var_dict["layers"]
            patches = None
            if "patches" in self.var_dict:
                patches = self.var_dict["patches"]
            accumulated = False
            if "accumulated" in self.var_dict:
                accumulated = self.var_dict["accumulated"]
            interval = None
            if "interval" in self.var_dict:
                interval = self.var_dict["interval"]
            datatype = None
            if "datatype" in self.var_dict:
                datatype = self.var_dict["datatype"]

            var = surfex.file.SurfexFileVariable(varname, validtime=validtime, patches=patches, layers=layers,
                                                 basetime=self.basetime,
                                                 interval=interval, datatype=datatype)
            int_type = "nearest"
            if "interpolator" in self.var_dict:
                int_type = self.var_dict["interpolator"]

            # Re-read field
            previous_field = None
            if accumulated:
                print(self.basetime, self.initialtime, self.previoustime)
                can_read = True
                if self.basetime <= self.initialtime:
                    if self.previoustime < self.basetime:
                        can_read = False

                if not can_read:
                    surfex.warning("Can not read previous time step for this time. Setting it to 0.")
                    print(self.basetime, self.initialtime, self.previoustime)
                    previous_field = np.zeros([geo.npoints])
                else:
                    id_str = None
                    if cache is not None:
                        id_str = cache.generate_surfex_id(varname, patches, layers, self.previousfilename,
                                                          self.previoustime)
                    if cache is not None and cache.is_saved(id_str):
                        previous_field = cache.saved_fields[id_str]
                    else:
                        fname = self.filename
                        if self.debug:
                            print("Re-read ", self.previoustime, " from ", self.previousfilename)
                        self.file_handler.fname = self.previousfilename
                        previous_field, intp = self.file_handler.points(var, geo, validime=self.previoustime,
                                                                        interpolation=int_type, cache=cache)

                        # Change filename back in handler. Ready to read this time step
                        self.file_handler.fname = fname
                        if cache is not None:
                            cache.save_field(id_str, previous_field)

            # Read field
            id_str = None
            if cache is not None:
                id_str = cache.generate_surfex_id(varname, patches, layers, self.filename, self.validtime)
            if cache is not None and cache.is_saved(id_str):
                field = cache.saved_fields[id_str]
            else:
                print(validtime)
                field, interpolator = self.file_handler.points(var, geo, validtime=validtime, interpolation=int_type,
                                                               cache=cache)
                # Rotate wind to geographic if requested
                field = self.rotate_geographic_wind(field, interpolator)
                if cache is not None:
                    cache.save_field(id_str, field)

            # Deaccumulate
            if accumulated:
                instant = [(validtime - self.previoustime).total_seconds()]
                if "instant" in self.var_dict:
                    instant = [self.var_dict["instant"]]
                field = self.deaccumulate(field, previous_field, float(instant[0]))

        self.previoustime = validtime
        return field

    def print_variable_info(self):
        print(":" + str(self.var_dict) + ":")


class FaVariable(Variable):

    """
    FA variable
    """

    def __init__(self, var_dict, basetime, validtime, debug):

        mandatory = ["name", "fcint", "offset", "file_inc", "filepattern"]
        for i in range(0, len(mandatory)):
            if mandatory[i] not in var_dict:
                raise Exception("NetCDF variable must have attribute " + mandatory[i] + " var_dict:" + str(var_dict))

        Variable.__init__(self, basetime, validtime, var_dict, debug)

    def read_variable(self, geo, validtime, cache, geo_in=None):

        self.validtime = validtime
        if self.open_new_file(int(self.var_dict["fcint"]), int(self.var_dict["offset"]),
                              int(self.var_dict["file_inc"])):
            # print "Updating filehandler for "+self.print_variable_info()
            if cache is not None and cache.file_open(self.filename):
                self.file_handler = cache.get_file_handler(self.filename)
            else:
                self.file_handler = surfex.fa.Fa(self.filename)
                if cache is not None:
                    cache.set_file_handler(self.filename, self.file_handler)

        if self.file_handler is None:
            surfex.util.warning("No file handler exist for this time step")
            field = np.array([len(geo.lons)])
            field = field.fill(np.nan)
        else:
            var_name = self.var_dict["name"]
            accumulated = False
            if "accumulated" in self.var_dict:
                accumulated = self.var_dict["accumulated"]
            int_type = "nearest"
            if "interpolator" in self.var_dict:
                int_type = self.var_dict["interpolator"]

            # Re-read field
            previous_field = None
            if accumulated:
                print(self.basetime, self.initialtime, self.previoustime)
                can_read = True
                if self.basetime <= self.initialtime:
                    if self.previoustime < self.basetime:
                        can_read = False

                if not can_read:
                    surfex.warning("Can not read previous time step for this time. Setting it to 0.")
                    print(self.basetime, self.initialtime, self.previoustime)
                    previous_field = np.zeros([geo.npoints])
                else:
                    # Re-read field
                    id_str = None
                    if cache is not None:
                        id_str = cache.generate_netcdf_id(var_name, self.previousfilename, self.previoustime)
                    if cache is not None and cache.is_saved(id_str):
                        print("Updating cached value ", id_str)
                        previous_field = cache.saved_fields[id_str]
                    else:
                        # Modify filename in handler
                        fname = self.filename
                        if self.debug:
                            print("Re-read ", self.previoustime, " from ", self.previousfilename)
                        self.file_handler.fname = self.previousfilename
                        previous_field, intp = self.file_handler.points(var_name,  geo,
                                                                        validtime=self.previoustime,
                                                                        interpolation=int_type)
                        if cache is not None:
                            cache.save_field(id_str, previous_field)
                        # Change filename back in handler. Ready to read this time step
                        self.file_handler.fname = fname

            field, interpolator = self.file_handler.points(var_name, geo, validtime=validtime,
                                                           interpolation=int_type, cache=cache)
            # Rotate wind to geographic if requested
            field = self.rotate_geographic_wind(field, interpolator)
            if cache is not None:
                id_str = cache.generate_netcdf_id(var_name, self.filename, validtime)
                cache.save_field(id_str, field)

            if accumulated:
                print("accumulated variable ", self.var_dict)
                print("field", field)
                print("prevous", previous_field)
                print("deccumulated", self.deaccumulate(field, previous_field, 0))
            if accumulated:
                instant = [(validtime - self.previoustime).total_seconds()]
                if "instant" in self.var_dict:
                    instant = [self.var_dict["instant"]]
                field = self.deaccumulate(field, previous_field, float(instant[0]))
                print("instant", field)

        self.previoustime = validtime
        return field

    def print_variable_info(self):
        print(":" + str(self.var_dict) + ":")


class ObservationVariable(Variable):
    def __init__(self,  var_dict, basetime, validtime, debug):

        mandatory = ["filetype", "fcint", "offset", "file_inc"]

        for i in range(0, len(mandatory)):
            print(mandatory[i], var_dict)
            if mandatory[i] not in var_dict:
                raise Exception("Observation variable must have attribute " + mandatory[i] + " var_dict:" +
                                str(var_dict))

        Variable.__init__(self, basetime, validtime, var_dict, debug)

    def read_variable(self, geo, validtime, cache, geo_in=None):

        self.validtime = validtime
        if self.open_new_file(int(self.var_dict["fcint"]), int(self.var_dict["offset"]),
                              int(self.var_dict["file_inc"])):

            if cache is not None and cache.file_open(self.filename):
                self.file_handler = cache.get_file_handler(self.filename)
            else:
                var_dict = self.var_dict
                var_dict = {"set": var_dict}
                self.file_handler = surfex.get_datasources(validtime, var_dict)[0]

                if cache is not None:
                    cache.set_file_handler(self.filename, self.file_handler)

        if self.file_handler is None:
            warning("No file handler exist for this time step")
            field = np.array([geo.nlons])
            field = field.fill(np.nan)
        else:
            if "varname" in self.var_dict:
                varname = self.var_dict["varname"]
            else:
                varname = None

            accumulated = False
            if "accumulated" in self.var_dict:
                accumulated = self.var_dict["accumulated"]

            # Re-read field
            previous_field = None
            if accumulated:
                print(self.basetime, self.initialtime, self.previoustime)
                can_read = True
                if self.basetime <= self.initialtime:
                    if self.previoustime < self.basetime:
                        can_read = False

                if not can_read:
                    surfex.warning("Can not read previous time step for this time. Setting it to 0.")
                    print(self.basetime, self.initialtime, self.previoustime)
                    previous_field = np.zeros([geo.npoints])
                else:
                    id_str = None
                    if cache is not None:
                        id_str = cache.generate_obs_id(varname, self.previousfilename, self.previoustime)
                    if cache is not None and cache.is_saved(id_str):
                        previous_field = cache.saved_fields[id_str]
                    else:
                        fname = self.filename
                        if self.debug:
                            print("Re-read ", self.previoustime, " from ", self.previousfilename)
                        self.file_handler.fname = self.previousfilename

                        times, previous_field, stids = self.file_handler.points(geo)

                        # Change filename back in handler. Ready to read this time step
                        self.file_handler.fname = fname
                        if cache is not None:
                            cache.save_field(id_str, previous_field)

            # Read field
            id_str = None
            if cache is not None:
                id_str = cache.generate_obs_id(varname, self.filename, self.validtime)
            if cache is not None and cache.is_saved(id_str):
                field = cache.saved_fields[id_str]
            else:
                times, field, stids = self.file_handler.points(geo)

            # Deaccumulate
            if accumulated:
                instant = [(validtime - self.previoustime).total_seconds()]
                if "instant" in self.var_dict:
                    instant = [self.var_dict["instant"]]
                field = self.deaccumulate(field, previous_field, float(instant[0]))

        self.previoustime = validtime
        return field

    def print_variable_info(self):
        pass



import surfex
import os
import subprocess
from abc import ABC, abstractmethod
import json
import shutil
import sys


class BatchJob(object):
    def __init__(self, rte, wrapper=""):
        self.rte = rte
        self.wrapper = wrapper
        print("Constructed BatchJob")

    def run(self, cmd):

        if cmd is None:
            raise Exception("No command provided!")
        cmd = self.wrapper + " " + cmd

        if "OMP_NUM_THREADS" in self.rte:
            print("BATCH: ", self.rte["OMP_NUM_THREADS"])
        print("Batch running " + cmd)

        process = subprocess.Popen(cmd, shell=True, env=self.rte, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                                   universal_newlines=True, bufsize=1)
        # Poll process for new output until finished
        while True:
            nextline = process.stdout.readline()
            if nextline == '' and process.poll() is not None:
                break
            sys.stdout.write(nextline)
            sys.stdout.flush()

        return_code = process.wait()
        if return_code != 0:
            raise subprocess.CalledProcessError(return_code, cmd)


class SURFEXBinary(object):
    def __init__(self, binary, batch, iofile, settings, input_data, **kwargs):
        self.binary = binary
        self.batch = batch
        self.iofile = iofile
        self.settings = settings

        self.surfout = None
        if "surfout" in kwargs:
            self.surfout = kwargs["surfout"]

        self.archive_data = None
        if "archive_data" in kwargs:
            self.archive_data = kwargs["archive_data"]

        self.print_namelist = False
        if "print_namelist" in kwargs:
            self.print_namelist = kwargs["print_namelist"]

        self.pgdfile = None
        if "pgdfile" in kwargs:
            self.pgdfile = kwargs["pgdfile"]

        # Set input
        self.input_data = input_data
        self.input_data.prepare_input()

        if os.path.exists('OPTIONS.nam'):
            os.remove('OPTIONS.nam')
        self.settings.write('OPTIONS.nam')
        fh = open('OPTIONS.nam')
        content = fh.read()
        fh.close()
        if self.print_namelist:
            print(content)

        if self.iofile.need_pgd and self.pgdfile is not None:
            print(self.pgdfile.filename)
            try:
                print("PGD is " + self.pgdfile.filename)
                if self.pgdfile.input_file is not None and \
                        os.path.abspath(self.pgdfile.filename) != os.path.abspath(self.pgdfile.input_file):
                    surfex.read.remove_existing_file(self.pgdfile.input_file, self.pgdfile.filename)
                    os.symlink(self.pgdfile.input_file, self.pgdfile.filename)
                if not os.path.exists(self.pgdfile.filename):
                    print("PGD not found! " + self.pgdfile.filename)
                    raise FileNotFoundError
            except FileNotFoundError:
                print("Could not set PGD")
                raise
            if self.surfout is not None:
                try:
                    print("PREP is " + self.iofile.filename)
                    if self.iofile.input_file is not None and \
                            os.path.abspath(self.iofile.filename) != os.path.abspath(self.iofile.input_file):
                        surfex.read.remove_existing_file(self.iofile.input_file, self.iofile.filename)
                        os.symlink(self.iofile.input_file, self.iofile.filename)
                    if not os.path.exists(self.iofile.filename):
                        raise FileNotFoundError("PREP not found! " + self.iofile.filename)
                except FileNotFoundError:
                    raise FileNotFoundError("Could not set PREP")

        cmd = self.binary
        print("Running " + cmd + " with settings OPTIONS.nam")
        try:
            self.batch.run(cmd)
        except Exception as e:
            raise RuntimeError(repr(e))

        listings = ["LISTING_PGD0.txt", "LISTING_PREP0.txt", "LISTING_OFFLINE0.txt", "LISTING_SODA0.txt"]
        for listing in listings:
            if os.path.exists(listing):
                fh = open(listing, "r")
                content = fh.read()
                fh.close()
                print("Content of " + listing + ":")
                print(content)

        # Archive output
        self.iofile.archive_output_file()
        if self.surfout is not None:
            self.surfout.archive_output_file()
        if self.archive_data is not None:
            self.archive_data.archive_files()


class PerturbedOffline(SURFEXBinary):
    def __init__(self, binary, batch, io, pert_number, settings, input_data, surfout=None,
                 archive_data=None, pgdfile=None, print_namelist=False):
        self.pert_number = pert_number
        settings['nam_io_varassim']['LPRT'] = True
        settings['nam_var']['nivar'] = int(pert_number)
        SURFEXBinary.__init__(self, binary, batch, io, settings, input_data, surfout=surfout,
                              archive_data=archive_data, pgdfile=pgdfile, print_namelist=print_namelist)


class Masterodb(object):
    def __init__(self, pgdfile, prepfile, surffile, settings, input_data, binary=None,
                 archive_data=None, print_namelist=True, batch=None):

        self.settings = settings
        self.binary = binary
        self.prepfile = prepfile
        self.surfout = surffile
        self.batch = batch
        self.pgdfile = pgdfile
        self.input = input_data
        self.archive = archive_data
        self.print_namelist = print_namelist

        # Set input
        self.input.prepare_input()

        # Prepare namelist
        if os.path.exists('EXSEG1.nam'):
            os.remove('EXSEG1.nam')

        self.settings.write('EXSEG1.nam')
        fh = open('EXSEG1.nam')
        content = fh.read()
        fh.close()
        if self.print_namelist:
            print(content)

        print("PGD file for MASTERODB", self.pgdfile.filename)
        if self.pgdfile.input_file is not None and \
                os.path.abspath(self.pgdfile.filename) != os.path.abspath(self.pgdfile.input_file):
            print("Input PGD file is: " + self.pgdfile.input_file)
            surfex.read.remove_existing_file(self.pgdfile.input_file, self.pgdfile.filename)
            os.symlink(self.pgdfile.input_file, self.pgdfile.filename)

        if not os.path.exists(self.pgdfile.filename):
            print("PGD not found! " + self.pgdfile.filename)
            raise FileNotFoundError

        print("PREP file for MASTERODB", self.prepfile.filename)
        if self.prepfile.input_file is not None and \
                os.path.abspath(self.prepfile.filename) != os.path.abspath(self.prepfile.input_file):

            print("Input PREP file is: " + self.prepfile.input_file)
            surfex.read.remove_existing_file(self.prepfile.input_file, self.prepfile.filename)
            os.symlink(self.prepfile.input_file, self.prepfile.filename)

        if not os.path.exists(self.prepfile.filename):
            print("PREP not found! " + self.prepfile.filename)
            raise FileNotFoundError

        # Archive if we have run the binary
        if self.binary is not None:
            print("Run binary with namelist EXSEG1.nam: " + self.binary)
            self.batch.run(self.binary)

    def archive_output(self):
        # Archive output
        self.surfout.archive_output_file()
        if self.archive is not None:
            self.archive.archive_files()


class InputDataToSurfexBinaries(ABC):

    def __init__(self):
        pass

    @abstractmethod
    def prepare_input(self):
        return NotImplementedError


class OutputDataFromSurfexBinaries(ABC):

    def __init__(self):
        pass

    @abstractmethod
    def archive_files(self):
        return NotImplementedError


class JsonOutputData(OutputDataFromSurfexBinaries):
    def __init__(self, data):
        OutputDataFromSurfexBinaries.__init__(self)
        self.data = data

    def archive_files(self):
        for output_file, target in self.data.items():

            print(output_file, target)
            command = "mv"
            if type(target) is dict:
                for key in target:
                    print(output_file, key, target[key])
                    command = target[key]
                    target = key

            cmd = command + " " + output_file + " " + target
            try:
                print(cmd)
                subprocess.check_call(cmd, shell=True)
            except IOError:
                print(cmd + " failed")
                raise


class JsonOutputDataFromFile(JsonOutputData):
    def __init__(self, file):
        JsonOutputData.__init__(self, json.load(open(file, "r")))

    def archive_files(self):
        JsonOutputData.archive_files(self)


class JsonInputData(InputDataToSurfexBinaries):
    def __init__(self, data):
        InputDataToSurfexBinaries.__init__(self)
        self.data = data

    def prepare_input(self):
        for target, input_file in self.data.items():

            print(target, input_file)
            print(os.path.realpath(target))
            command = None
            if type(input_file) is dict:
                for key in input_file:
                    print(key)
                    print(input_file[key])
                    command = str(input_file[key])
                    input_file = str(key)
                    command = command.replace("@INPUT@", input_file)
                    command = command.replace("@TARGET@", target)

            if os.path.realpath(target) == os.path.realpath(input_file):
                print("Target and input file is the same file")
            else:
                if command is None:
                    cmd = "ln -sf " + input_file + " " + target
                else:
                    cmd = command
                try:
                    print(cmd)
                    subprocess.check_call(cmd, shell=True)
                except IOError:
                    raise(cmd + " failed")

    def add_data(self, data):
        for key in data:
            value = data[key]
            self.data.update({key: value})


class JsonInputDataFromFile(JsonInputData):
    def __init__(self, file):
        JsonInputData.__init__(self, json.load(open(file, "r")))

    def prepare_input(self):
        JsonInputData.prepare_input(self)


def create_working_dir(workdir, enter=True):
    # Create work directory
    if workdir is not None:
        if os.path.isdir(workdir):
            shutil.rmtree(workdir)
        os.makedirs(workdir, exist_ok=True)
        if enter:
            os.chdir(workdir)


def clean_working_dir(workdir):
    if exit:
        os.chdir("..")
    # Clean up
    shutil.rmtree(workdir)



import abc
import netCDF4
import numpy as np
import time
import copy
import shutil
import surfex
from datetime import datetime, timedelta


class SurfexForcing(object):
    """
    Main output class for SURFEX forcing
    """

    __metaclass__ = abc.ABCMeta

    def __init__(self, base_time, geo, ntimes, var_objs, debug):
        self.time_step_intervall = 3600
        self.valid_time = None
        self.base_time = base_time
        self.geo = geo
        self.debug = debug
        self.ntimes = ntimes
        self.time_step = 0
        self.var_objs = var_objs
        self._check_sanity()

    def _check_sanity(self):
        if len(self.var_objs) != self.nparameters:
            print("Inconsistent number of parameter. " + str(len(self.var_objs)) + " != " + str(self.nparameters))
            raise Exception

        # Check if all parameters are present
        for i in range(0, len(self.var_objs)):
            self.parameters[self.var_objs[i].var_name] = 1

        for key in self.parameters:
            if self.parameters[key] == 0:
                print("Required parameter " + str(key) + " is missing!")
                raise Exception

    # Time dependent parameter
    nparameters = 11
    parameters = {
        "TA": 0,
        "QA": 0,
        "PS": 0,
        "DIR_SW": 0,
        "SCA_SW": 0,
        "LW": 0,
        "RAIN": 0,
        "SNOW": 0,
        "WIND": 0,
        "WIND_DIR": 0,
        "CO2": 0,
    }

    @abc.abstractmethod
    def write_forcing(self, var_objs, this_time, cache):
        raise NotImplementedError('users must define writeForcing to use this base class')


class NetCDFOutput(SurfexForcing):
    """

    Forcing in NetCDF format

    """

    output_format = "NETCDF3_64BIT"
    forcing_file = {}

    translation = {
        "TA": "Tair",
        "QA": "Qair",
        "PS": "PSurf",
        "DIR_SW": "DIR_SWdown",
        "SCA_SW": "SCA_SWdown",
        "LW": "LWdown",
        "RAIN": "Rainf",
        "SNOW": "Snowf",
        "WIND": "Wind",
        "WIND_DIR": "Wind_DIR",
        "CO2": "CO2air",
    }

    def __init__(self, base_time, geo, fname, ntimes, var_objs, att_objs, att_time, cache):
        SurfexForcing.__init__(self, base_time, geo, ntimes, var_objs, cache.debug)
        print("Forcing type is netCDF")
        self.forcing_file = {}
        if fname is None:
            fname = "FORCING.nc"
        self.fname = fname
        self.tmp_fname = self.fname + ".tmp"
        self.file_handler = netCDF4.Dataset(self.tmp_fname, 'w', format=self.output_format)
        self._define_forcing(geo, att_objs, att_time, cache)

    def write_forcing(self, var_objs, this_time, cache):

        # VARS
        for i in range(0, len(self.var_objs)):
            this_obj = self.var_objs[i]
            this_var = this_obj.var_name
            print(this_obj.var_name)
            tic = time.time()
            field = this_obj.read_time_step(this_time, cache)
            toc = time.time()
            print("# read_time_step: ", toc - tic)
            self.forcing_file[self.translation[this_var]][self.time_step, :] = field

        self.forcing_file['TIME'][self.time_step] = self.time_step

    def _define_forcing(self, geo, att_objs, att_time, cache):
        print("Define netcdf forcing")

        zs = None
        zref = None
        uref = None
        for i in range(0, len(att_objs)):
            this_obj = att_objs[i]
            this_var = this_obj.var_name
            print(this_obj.var_name)
            if this_var == "ZS":
                zs = this_obj.read_time_step(att_time, cache)
            elif this_var == "ZREF":
                zref = this_obj.read_time_step(att_time, cache)
            elif this_var == "UREF":
                uref = this_obj.read_time_step(att_time, cache)

        # DIMS
        self.forcing_file['NPOINTS'] = self.file_handler.createDimension("Number_of_points", geo.npoints)
        self.forcing_file['NTIMES'] = self.file_handler.createDimension("time", self.ntimes)

        # DEFINE VARS
        self.forcing_file['TIME'] = self.file_handler.createVariable("time", "f4", ("time",))
        self.forcing_file['TIME'].units = "hours since %s:00:00 0:00" % self.base_time.strftime("%Y-%m-%d %H")
        self.forcing_file['TSTEP'] = self.file_handler.createVariable("FRC_TIME_STP", "f4")
        self.forcing_file['TSTEP'].longname = "Forcing_Time_Step"
        self.forcing_file['TSTEP'][:] = self.time_step_intervall
        self.forcing_file['LON'] = self.file_handler.createVariable("LON", "f4", ("Number_of_points",))
        self.forcing_file['LON'].longname = "Longitude"
        self.forcing_file['LON'][:] = geo.lonlist
        self.forcing_file['LAT'] = self.file_handler.createVariable("LAT", "f4", ("Number_of_points",))
        self.forcing_file['LAT'].longname = "Latitude"
        self.forcing_file['LAT'][:] = geo.latlist
        self.forcing_file['ZS'] = self.file_handler.createVariable("ZS", "f4", ("Number_of_points",))
        self.forcing_file['ZS'].longname = "Surface_Orography"
        self.forcing_file['ZS'][:] = zs
        self.forcing_file['ZREF'] = self.file_handler.createVariable("ZREF", "f4", ("Number_of_points",))
        self.forcing_file['ZREF'].longname = "Reference_Height"
        self.forcing_file['ZREF'].units = "m"
        self.forcing_file['ZREF'][:] = zref
        self.forcing_file['UREF'] = self.file_handler.createVariable("UREF", "f4", ("Number_of_points",))
        self.forcing_file['UREF'].longname = "Reference_Height_for_Wind"
        self.forcing_file['UREF'].units = "m"
        self.forcing_file['UREF'][:] = uref

        # Define time dependent variables
        for i in range(0, len(self.var_objs)):
            this_obj = self.var_objs[i]
            this_var = this_obj.var_name

            # print this_var
            if this_var == "TA":
                self.forcing_file['Tair'] = self.file_handler.createVariable("Tair", "f4",
                                                                             ("time", "Number_of_points",))
                self.forcing_file['Tair'].longname = "Near_Surface_Air_Temperature"
                self.forcing_file['Tair'].units = "K"
            elif this_var == "QA":
                self.forcing_file['Qair'] = self.file_handler.createVariable("Qair", "f4", ("time",
                                                                                            "Number_of_points",))
                self.forcing_file['Qair'].longname = "Near_Surface_Specific_Humidity"
                self.forcing_file['Qair'].units = "kg/kg"
            elif this_var == "PS":
                self.forcing_file['PSurf'] = self.file_handler.createVariable("PSurf", "f4", ("time",
                                                                                              "Number_of_points",))
                self.forcing_file['PSurf'].longname = "Surface_Pressure"
                self.forcing_file['PSurf'].units = "Pa"
            elif this_var == "DIR_SW":
                self.forcing_file['DIR_SWdown'] = self.file_handler.createVariable("DIR_SWdown", "f4",
                                                                                   ("time", "Number_of_points",))
                self.forcing_file['DIR_SWdown'].longname = "Surface_Incident_Downwelling_Shortwave_Radiation"
                self.forcing_file['DIR_SWdown'].units = "W/m2"
            elif this_var == "SCA_SW":
                self.forcing_file['SCA_SWdown'] = self.file_handler.createVariable("SCA_SWdown", "f4",
                                                                                   ("time", "Number_of_points",))
                self.forcing_file['SCA_SWdown'].longname = "Surface_Incident_Diffuse_Shortwave_Radiation"
                self.forcing_file['SCA_SWdown'].units = "W/m2"
            elif this_var == "LW":
                self.forcing_file['LWdown'] = self.file_handler.createVariable("LWdown", "f4", ("time",
                                                                                                "Number_of_points",))
                self.forcing_file['LWdown'].longname = "Surface_Incident_Diffuse_Longwave_Radiation"
                self.forcing_file['LWdown'].units = "W/m2"
            elif this_var == "RAIN":
                self.forcing_file['Rainf'] = self.file_handler.createVariable("Rainf", "f4", ("time",
                                                                                              "Number_of_points",))
                self.forcing_file['Rainf'].longname = "Rainfall_Rate"
                self.forcing_file['Rainf'].units = "kg/m2/s"
            elif this_var == "SNOW":
                self.forcing_file['Snowf'] = self.file_handler.createVariable("Snowf", "f4", ("time",
                                                                                              "Number_of_points",))
                self.forcing_file['Snowf'].longname = "Snowfall_Rate"
                self.forcing_file['Snowf'].units = "kg/m2/s"
            elif this_var == "WIND":
                self.forcing_file['Wind'] = self.file_handler.createVariable("Wind", "f4", ("time",
                                                                                            "Number_of_points",))
                self.forcing_file['Wind'].longname = "Wind_Speed"
                self.forcing_file['Wind'].units = "m/s"
            elif this_var == "WIND_DIR":
                self.forcing_file['Wind_DIR'] = self.file_handler.createVariable("Wind_DIR", "f4",
                                                                                 ("time", "Number_of_points",))
                self.forcing_file['Wind_DIR'].longname = "Wind_Direction"
            elif this_var == "CO2":
                self.forcing_file['CO2air'] = self.file_handler.createVariable("CO2air", "f4", ("time",
                                                                                                "Number_of_points",))
                self.forcing_file['CO2air'].longname = "Near_Surface_CO2_Concentration"
                self.forcing_file['CO2air'].units = "kg/m3"
            else:
                print("This should never happen! " + this_var + " is not defined!")
                raise NotImplementedError

    def finalize(self):
        print("Close file")
        self.file_handler.close()
        shutil.move(self.tmp_fname, self.fname)


class AsciiOutput(SurfexForcing):
    """

    Forcing in ASCII format

    """

    output_format = "ascii"

    def __init__(self, base_time, geo, fname, ntimes, var_objs, att_objs, att_time, cache):
        SurfexForcing.__init__(self, base_time, geo, ntimes, var_objs, cache.debug)
        print("Forcing type is ASCII")
        self.forcing_file = {}
        self.file_handler = {}
        if fname is None:
            fname = "Params_config.txt"
        self.fname = fname
        self._define_forcing(geo, att_objs, att_time, cache)

    def write_forcing(self, var_objs, this_time, cache):
        for i in range(0, len(self.var_objs)):
            this_obj = self.var_objs[i]
            this_var = this_obj.var_name
            print(this_obj.var_name)
            field = this_obj.read_time_step(this_time, cache)
            fmt = "%20.8f"
            cols = 50
            write_formatted_array(self.file_handler[this_var], field, cols, fmt)

    def _define_forcing(self, geo, att_objs, att_time, cache):
        zs = None
        zref = None
        uref = None
        for i in range(0, len(att_objs)):
            this_obj = att_objs[i]
            this_var = this_obj.var_name
            if this_var == "ZS":
                zs = this_obj.read_time_step(att_time, cache)
            elif this_var == "ZREF":
                zref = this_obj.read_time_step(att_time, cache)
            elif this_var == "UREF":
                uref = this_obj.read_time_step(att_time, cache)

        second = (self.base_time - self.base_time.replace(hour=0,
                                                          minute=0,
                                                          second=0,
                                                          microsecond=0)).total_seconds()
        fmt = "%15.8f"
        cols = 50
        f = open(self.fname, 'w')
        f.write(str(geo.npoints) + '\n')
        f.write(str(self.ntimes) + '\n')
        f.write(str(self.time_step_intervall) + '\n')
        f.write(self.base_time.strftime("%Y") + '\n')
        f.write(self.base_time.strftime("%m") + '\n')
        f.write(self.base_time.strftime("%d") + '\n')
        f.write(str(second) + '\n')
        write_formatted_array(f, geo.lons, cols, fmt)
        write_formatted_array(f, geo.lats, cols, fmt)
        write_formatted_array(f, zs, cols, fmt)
        write_formatted_array(f, zref, cols, fmt)
        write_formatted_array(f, uref, cols, fmt)
        f.close()

        for key in self.parameters:
            nam = key
            if key == "WIND_DIR":
                nam = "DIR"
            self.forcing_file[key] = "Forc_" + nam + '.txt'
            self.file_handler[key] = open(self.forcing_file[key], 'w')

    def finalize(self):
        print("Close file")
        for key in self.parameters:
            self.file_handler[key].close()


def write_formatted_array(file, array, columns, fileformat):
    astr = np.empty(array.size - array.size % columns, dtype="float64")
    astr = array[0:astr.size]
    astr = astr.reshape((columns, astr.size / columns), order='F')
    mlw = (len(fileformat % 0)) * (columns + 1)
    formatter = {'float_kind': lambda x: fileformat % x}
    astr_end = np.array2string(array[astr.size:],
                               separator='',
                               max_line_width=mlw,
                               formatter=formatter)[1:-1]
    np.savetxt(file, astr.T, fmt=fileformat, newline='\n', delimiter='')
    file.write(astr_end + '\n')


def run_time_loop(options, var_objs, att_objs):

    this_time = options['start']
    cache = surfex.cache.Cache(options['debug'], options['cache_interval'])
    # Find how many time steps we want to write
    ntimes = 0
    while this_time <= options['stop']:
        ntimes = ntimes+1
        this_time = this_time + timedelta(seconds=options['timestep'])

    # Create output object
    if str.lower(options['output_format']) == "netcdf":
        # Set att_time the same as start
        att_time = options['start']
        output = surfex.forcing.NetCDFOutput(options['start'], options['geo_out'], options['output_file'], ntimes,
                                             var_objs, att_objs, att_time, cache)
    elif str.lower(options['output_format']) == "ascii":
        att_time = options['start']
        # base_time, geo, ntimes, var_objs, att_objs, att_time, cache
        output = surfex.forcing.AsciiOutput(options['start'], options['geo_out'], options['output_file'], ntimes,
                                            var_objs, att_objs, att_time, cache)
    else:
        print("Invalid output format "+options['output_format'])
        raise NotImplementedError

    # Loop output time steps
    this_time = options['start']
    while this_time <= options['stop']:

        # Write for each time step
        print("Creating forcing for: " + this_time.strftime('%Y%m%d%H') + " time_step:" + str(output.time_step))
        output.write_forcing(var_objs, this_time, cache)
        output.time_step = output.time_step + 1
        this_time = this_time + timedelta(seconds=options['timestep'])
        cache.clean_fields(this_time)

    # Finalize forcing
    output.finalize()


def set_input_object(sfx_var, merged_conf, geo, forcingformat, selected_converter, ref_height, start, first_base_time,
                     timestep, debug):
    """
    Set the input parameter for a specific SURFEX forcing variable based on input

    :param sfx_var:
    :param merged_conf:
    :param geo:
    :param forcingformat:
    :param selected_converter:
    :param ref_height:
    :param start:
    :param first_base_time:
    :param timestep:
    :param debug:
    :return:
    """

    #########################################
    # 1. Gobal configuration from yaml file
    #########################################

    conf = copy.deepcopy(merged_conf)

    # Now we know the CLA settings for each surfex variable

    # Set defaults (merged global and user settings)
    # Theses must be sent to the converter to be used for each variable
    defs = {}
    if forcingformat in conf:
        defs = copy.deepcopy(conf[forcingformat])
        defs.update({"timestep": timestep})

    # All objects with converters, find converter dict entry
    conf_dict = {}
    if forcingformat != "constant":

        # Non-height dependent variables
        if ref_height is None:
            if "converter" in conf[sfx_var][forcingformat]:
                conf_dict = copy.deepcopy(conf[sfx_var][forcingformat]["converter"])
            else:
                print("No converter defined for " + sfx_var)
                raise KeyError

        # Variables with height dependency
        else:
            if ref_height in conf[sfx_var]:
                # print sfx_var,ref_height,format
                if forcingformat not in conf[sfx_var][ref_height]:
                    print(str(conf[sfx_var]) + "\n Missing definitions for " + sfx_var + " and format: " +
                          forcingformat)
                    raise KeyError
                if conf[sfx_var][ref_height][forcingformat] is None:
                    print(str(conf[sfx_var]) + "\n Missing definitions for " + sfx_var)
                    raise KeyError
                if "converter" in conf[sfx_var][ref_height][forcingformat]:
                    conf_dict = copy.deepcopy(conf[sfx_var][ref_height][forcingformat]["converter"])
                else:
                    print("No converter defined for " + sfx_var)
                    raise KeyError
            else:
                print("No ref height \"" + ref_height + "\" defined for " + sfx_var)
                raise KeyError

    ##############################################################
    ##############################################################
    ##############################################################
    # Create the object to be returned
    if forcingformat == "constant":
        if ref_height is None:
            if "constant" in conf[sfx_var]:
                const_dict = copy.deepcopy(conf[sfx_var]["constant"])
            else:
                print("No constant defined for " + sfx_var)
                raise KeyError
        else:
            if ref_height in conf[sfx_var]:
                if "constant" in conf[sfx_var][ref_height]:
                    const_dict = copy.deepcopy(conf[sfx_var][ref_height]["constant"])
                else:
                    print("No constant defined for " + sfx_var)
                    raise KeyError
            else:
                print("No ref height \"" + ref_height + "\" defined for " + sfx_var)
                raise KeyError

        obj = surfex.read.ConstantValue(geo, sfx_var, const_dict)
    else:

        # Construct the converter
        # print sfx_var
        converter = surfex.read.Converter(selected_converter, start, defs, conf_dict, forcingformat,
                                          first_base_time, debug)

        # Construct the input object
        obj = surfex.read.ConvertedInput(geo, sfx_var, converter)
    return obj


def set_forcing_config(**kwargs):

    debug = False
    fb = None
    geo_out = None
    user_config = {}
    pattern = None
    timestep = 3600
    cache_interval = 3600
    zsoro = "default"
    zsoro_converter = "none"
    zval = "default"
    zval_converter = "none"
    uval = "default"
    uval_converter = "none"

    ta = "default"
    ta_converter = "none"
    qa = "default"
    qa_converter = "none"
    ps = "default"
    ps_converter = "none"
    dir_sw = "default"
    dir_sw_converter = "none"
    sca_sw = "default"
    sca_sw_converter = "none"
    lw = "default"
    lw_converter = "none"
    rain = "default"
    rain_converter = "none"
    snow = "default"
    snow_converter = "none"
    wind = "default"
    wind_converter = "none"
    wind_dir = "default"
    wind_dir_converter = "none"
    co2 = "default"
    co2_converter = "none"

    try:
        dtg_start = kwargs["dtg_start"]
        dtg_stop = kwargs["dtg_stop"]
        input_format = kwargs["input_format"]
        output_format = kwargs["output_format"]
        of = kwargs["of"]
        zref = kwargs["zref"]
        uref = kwargs["uref"]
        config = kwargs["config"]
        if "debug" in kwargs:
            debug = kwargs["debug"]
        if "fb" in kwargs:
            fb = kwargs["fb"]
        if "geo_out" in kwargs:
            geo_out = kwargs["geo_out"]
        if "user_config" in kwargs:
            user_config = kwargs["user_config"]
        if "pattern" in kwargs:
            pattern = kwargs["pattern"]
        if "timestep" in kwargs:
            timestep = kwargs["timestep"]
        if "cache_interval" in kwargs:
            cache_interval = kwargs["cache_interval"]
        if "zsoro" in kwargs:
            zsoro = kwargs["zsoro"]
        if "zsoro_converter" in kwargs:
            zsoro_converter = kwargs["zsoro_converter"]
        if "zval" in kwargs:
            zval = kwargs["zval"]
        if "zval_converter" in kwargs:
            zval_converter = kwargs["zval_converter"]
        if "uval_converter" in kwargs:
            uval_converter = kwargs["uval_converter"]
        if "uval" in kwargs:
            uval = kwargs["uval"]
        if "ta" in kwargs:
            ta = kwargs["ta"]
        if "ta_converter" in kwargs:
            ta_converter = kwargs["ta_converter"]
        if "qa" in kwargs:
            qa = kwargs["qa"]
        if "qa_converter" in kwargs:
            qa_converter = kwargs["qa_converter"]
        if "ps" in kwargs:
            ps = kwargs["ps"]
        if "ps_converter" in kwargs:
            ps_converter = kwargs["ps_converter"]
        if "dir_sw" in kwargs:
            dir_sw = kwargs["dir_sw"]
        if "dir_sw_converter" in kwargs:
            dir_sw_converter = kwargs["dir_sw_converter"]
        if "sca_sw" in kwargs:
            sca_sw = kwargs["sca_sw"]
        if "sca_sw_converter" in kwargs:
            sca_sw_converter = kwargs["sca_sw_converter"]
        if "lw" in kwargs:
            lw = kwargs["lw"]
        if "lw_converter" in kwargs:
            lw_converter = kwargs["lw_converter"]
        if "rain" in kwargs:
            rain = kwargs["rain"]
        if "rain_converter" in kwargs:
            rain_converter = kwargs["rain_converter"]
        if "snow" in kwargs:
            snow = kwargs["snow"]
        if "snow_converter" in kwargs:
            snow_converter = kwargs["snow_converter"]
        if "wind" in kwargs:
            wind = kwargs["wind"]
        if "wind_converter" in kwargs:
            wind_converter = kwargs["wind_converter"]
        if "wind_dir" in kwargs:
            wind_dir = kwargs["wind_dir"]
        if "wind_dir_converter" in kwargs:
            wind_dir_converter = kwargs["wind_dir_converter"]
        if "co2" in kwargs:
            co2 = kwargs["co2"]
        if "co2_converter" in kwargs:
            co2_converter = kwargs["co2_converter"]

    except ValueError:
        raise Exception("Needed input is missing")

    # Time information
    if (int(dtg_start) or int(dtg_stop)) < 1000010100:
        print("Invalid start and stop times! " + str(dtg_start) + " " + str(dtg_stop))
        raise Exception

    start = datetime.strptime(str.strip(str(dtg_start)), '%Y%m%d%H')
    stop = datetime.strptime(str.strip(str(dtg_stop)), '%Y%m%d%H')
    if fb is None:
        first_base_time = start
    else:
        first_base_time = datetime.strptime(str.strip(str(fb)), '%Y%m%d%H')

    # Merge all settings with user all settings
    merged_conf = surfex.util.data_merge(config, user_config)

    # Replace global settings from
    fileformat = input_format
    if pattern is not None:
        merged_conf[fileformat]["filepattern"] = pattern

    # Set attributes
    atts = ["ZS", "ZREF", "UREF"]
    att_objs = []
    for i in range(0, len(atts)):

        att_var = atts[i]
        # Override with command line options for a given variable
        ref_height = None
        cformat = fileformat
        if att_var == "ZS":
            if zsoro != "default":
                cformat = zsoro
            selected_converter = zsoro_converter
        elif att_var == "ZREF":
            if zval != "default":
                cformat = zval
            selected_converter = zval_converter
            ref_height = zref
            if ref_height == "screen":
                cformat = "constant"
        elif att_var == "UREF":
            if uval != "default":
                cformat = uval
            selected_converter = uval_converter
            ref_height = uref
            if ref_height == "screen":
                cformat = "constant"
        else:
            raise NotImplementedError

        att_objs.append(set_input_object(atts[i], merged_conf, geo_out, cformat, selected_converter, ref_height,
                                         start, first_base_time, timestep, debug))

    # Set forcing variables (time dependent)
    variables = ["TA", "QA", "PS", "DIR_SW", "SCA_SW", "LW", "RAIN", "SNOW", "WIND", "WIND_DIR", "CO2"]
    var_objs = []
    # Search in config file for parameters to override
    for i in range(0, len(variables)):

        ref_height = None
        sfx_var = variables[i]
        cformat = fileformat
        if sfx_var == "TA":
            if ta != "default":
                cformat = ta
            selected_converter = ta_converter
            ref_height = zref
        elif sfx_var == "QA":
            if qa != "default":
                cformat = qa
            selected_converter = qa_converter
            ref_height = zref
        elif sfx_var == "PS":
            if ps != "default":
                cformat = ps
            selected_converter = ps_converter
        elif sfx_var == "DIR_SW":
            if dir_sw != "default":
                cformat = dir_sw
            selected_converter = dir_sw_converter
        elif sfx_var == "SCA_SW":
            if sca_sw != "default":
                cformat = sca_sw
            selected_converter = sca_sw_converter
        elif sfx_var == "LW":
            if lw != "default":
                cformat = lw
            selected_converter = lw_converter
        elif sfx_var == "RAIN":
            if rain != "default":
                cformat = rain
            selected_converter = rain_converter
        elif sfx_var == "SNOW":
            if snow != "default":
                cformat = snow
            selected_converter = snow_converter
        elif sfx_var == "WIND":
            if wind != "default":
                cformat = wind
            selected_converter = wind_converter
            ref_height = uref
        elif sfx_var == "WIND_DIR":
            if wind_dir != "default":
                cformat = wind_dir
            selected_converter = wind_dir_converter
            ref_height = uref
        elif sfx_var == "CO2":
            if co2 != "default":
                cformat = co2
            selected_converter = co2_converter
        else:
            raise NotImplementedError
        var_objs.append(set_input_object(sfx_var, merged_conf, geo_out, cformat, selected_converter, ref_height,
                                         start, first_base_time, timestep, debug))

    # Save options
    options = dict()
    options['output_format'] = output_format
    options['output_file'] = of
    options['start'] = start
    options['stop'] = stop
    options['timestep'] = timestep
    options['geo_out'] = geo_out
    options['debug'] = debug
    options['cache_interval'] = cache_interval

    return options, var_objs, att_objs



import unittest
import surfex


class GridppTest(unittest.TestCase):

    def setUp(self):
        self.testdata = "testdata/"

    def test_gridpp_t2m(self):
        argv = ["-i", self.testdata + "/unittest_FirstGuess4gridpp_grib2.nc",
                "-o", "unittest_an_t2m.nc",
                "-obs", self.testdata + "/unittest_qc_t2m.json",
                "-hor", "30000",
                "-vert", "300",
                "-v", "air_temperature_2m",
                "--elevGradient", "-0.0065"
                ]
        kwargs = surfex.parse_args_gridpp(argv)
        print(kwargs)
        surfex.run_gridpp(**kwargs)

    def test_gridpp_rh2m(self):
        argv = ["-i", self.testdata + "/unittest_FirstGuess4gridpp_grib1.nc",
                "-o", "unittest_an_rh2m.nc",
                "-obs", self.testdata + "/unittest_qc_rh2m.json",
                "-hor", "30000",
                "-vert", "600",
                "-v", "relative_humidity_2m",
                "--elevGradient", "0"
                ]
        kwargs = surfex.parse_args_gridpp(argv)
        print(kwargs)
        surfex.run_gridpp(**kwargs)

    def test_gridpp_sd(self):
        argv = ["-i", self.testdata + "/unittest_FirstGuess4gridpp_grib2.nc",
                "-o", "unittest_an_sd.nc",
                "-obs", self.testdata + "/unittest_qc_sd.json",
                "-hor", "60000",
                "-vert", "500",
                "-v", "surface_snow_thickness",
                "--elevGradient", "0"
                ]
        kwargs = surfex.parse_args_gridpp(argv)
        print(kwargs)
        surfex.run_gridpp(**kwargs)



__version__ = '0.0.1a10'

from .file import *
from .geo import *
from .read import *
from .run import *
from .namelist import *
from .assim import *
from .obs import *
from .forcing import *
from .netcdf import *
from .grib import *
from .bufr import *
from .variable import *
from .util import *
from .cache import *
from .interpolation import *
from .titan import *
from .timeseries import *
from .obsmon import *
from .fa import *
from .cli import *
from .configuration import *



import unittest
import surfex
import sqlite3


class ObsmonTest(unittest.TestCase):

    def setUp(self):
        self.testdata = "testdata/"

    def test_obsmon(self):
        db = "unittest_ecma.db"

        argv = ["2020111306",
                "t2m",
                self.testdata + "/unittest_qc_t2m.json",
                "--fg_file", self.testdata + "/unittest_FirstGuess4gridpp_grib2.nc",
                "--an_file", self.testdata + "/unittest_an_t2m.nc",
                "--file_var", "air_temperature_2m",
                "-o", db
                ]
        kwargs = surfex.parse_args_qc2obsmon(argv)
        print(kwargs)
        surfex.write_obsmon_sqlite_file(**kwargs)

        sqlite3.connect(db)



import numpy as np
try:
    import gridpp
except ImportError:
    gridpp = None


def horizontal_oi(geo, background, observations, gelevs, glafs, hlength=10000.,
                  vlength=10000., wlength=0.5, elev_gradient=0, structure_function="Barnes",
                  max_locations=50, epsilon=0.5, minvalue=None, maxvalue=None,
                  interpol="bilinear", debug=False):

    if gridpp is None:
        raise Exception("You need gridpp to perform OI")

    if debug:
        print(gridpp.__file__)
        print(gridpp.__version__)
    glats = geo.lats
    glons = geo.lons

    def obs2vectors(my_obs):
        return my_obs.lons, my_obs.lats, my_obs.stids, my_obs.elevs, \
               my_obs.values, my_obs.cis, my_obs.lafs

    vectors = np.vectorize(obs2vectors)
    lons, lats, stids, elevs, values, pci, lafs = vectors(observations)

    glats = np.transpose(glats)
    glons = np.transpose(glons)
    background = np.transpose(background)
    gelevs = np.transpose(gelevs)

    bgrid = gridpp.Grid(glats, glons, gelevs)
    points = gridpp.Points(lats, lons, elevs)
    if interpol == "bilinear":
        pbackground = gridpp.bilinear(bgrid, points, background)
    elif interpol == "nearest":
        pbackground = gridpp.nearest(bgrid, points, background, elev_gradient)
    else:
        raise NotImplementedError
    variance_ratios = np.full(points.size(), epsilon)

    if structure_function == "Barnes":
        structure = gridpp.BarnesStructure(hlength, vlength, wlength)
    else:
        raise NotImplementedError

    field = gridpp.optimal_interpolation(bgrid, background, points, values, variance_ratios, pbackground, structure,
                                         max_locations)
    field = np.asarray(field)
    if minvalue is not None:
        field[field < minvalue] = minvalue
    if maxvalue is not None:
        field[field > maxvalue] = maxvalue
    return np.transpose(field)



import os
import surfex
import shutil
import numpy as np
from datetime import timedelta, datetime
# from netCDF4 import Dataset, num2date, chartostring
import netCDF4
import re
import abc
import pyproj
print(netCDF4.__file__)

class SurfexIO(object):
    def __init__(self, filename, geo, extension):
        self.filename = filename
        self.geo = geo
        self.extension = extension

    @abc.abstractmethod
    def field(self, var, validtime=None):
        raise NotImplementedError("This method is not implemented for this class!")

    @abc.abstractmethod
    def points(self, var, geo_out, validtime=None, interpolation="nearest", cache=None):
        raise NotImplementedError("This method is not implemented for this class!")

    def interpolate_field(self, field, geo_in, geo_out, interpolation="nearest", cache=None):

        if interpolation == "nearest":
            surfex.util.info("Nearest neighbour", level=2)
            interpolator = surfex.interpolation.NearestNeighbour(geo_in, geo_out, cache=cache)
        elif interpolation == "linear":
            surfex.util.info("Linear interpolation", level=2)
            interpolator = surfex.interpolation.Linear(geo_in, geo_out, cache=cache)
        elif interpolation == "none":
            surfex.util.info("No interpolation", level=2)
            interpolator = surfex.interpolation.NoInterpolation(geo_in, geo_out, cache=cache)
        else:
            raise NotImplementedError("Interpolation type " + interpolation + " not implemented!")

        field = interpolator.interpolate(field)
        return field, interpolator


class SurfexSurfIO(object):
    def __init__(self, surfexfile, csurf_filetype, input_file=None, symlink=True, archive_file=None):
        self.filename = surfexfile.filename
        self.csurf_filetype = csurf_filetype
        self.need_pgd = False
        self.input_file = input_file
        self.symlink = symlink
        if self.input_file is not None:
            if self.symlink:
                self.symlink_input()
            else:
                self.copy_input()
        self.archive_file = archive_file

    def symlink_input(self):
        if self.input_file is not None:
            f_out = os.getcwd() + "/" + self.filename
            surfex.read.remove_existing_file(self.input_file, f_out)
            if os.path.abspath(self.input_file) != f_out:
                print("Symlink " + self.input_file + " -> " + f_out)
                os.symlink(self.input_file, f_out)

    def copy_input(self):
        if self.input_file is not None:
            f_out = os.getcwd() + "/" + self.filename
            surfex.read.remove_existing_file(self.input_file, f_out)
            if os.path.abspath(self.input_file) != f_out:
                print("Copy " + self.input_file + " -> " + f_out)
                shutil.copy2(self.input_file, f_out)

    def archive_output_file(self):
        if self.archive_file is not None:
            dirname = os.path.dirname(os.path.abspath(self.archive_file))
            os.makedirs(dirname, exist_ok=True)
            f_in = os.getcwd() + "/" + self.filename
            if os.path.abspath(self.archive_file) != f_in:
                print("Move " + f_in + " to " + self.archive_file)
                if os.path.islink(self.archive_file):
                    # print("is link")
                    os.unlink(self.archive_file)
                if os.path.isfile(self.archive_file):
                    # print("is file")
                    os.remove(self.archive_file)
                if os.path.isdir(self.archive_file):
                    shutil.rmtree(self.archive_file)
                shutil.move(f_in, self.archive_file)


class PGDFile(SurfexSurfIO):
    def __init__(self, csurf_filetype, cpgdfile, geo, input_file=None, symlink=True, archive_file=None,
                 lfagmap=False, masterodb=False):
        print(cpgdfile, csurf_filetype, masterodb)

        cpgdfile = get_surfex_io_object(cpgdfile, filetype="surf", geo=geo, fileformat=csurf_filetype,
                                        lfagmap=lfagmap, masterodb=masterodb)

        SurfexSurfIO.__init__(self, cpgdfile, csurf_filetype, input_file=input_file,
                              archive_file=archive_file, symlink=symlink)
        self.need_pgd = False
        

class PREPFile(SurfexSurfIO):
    def __init__(self, csurf_filetype, cprepfile, geo, input_file=None, symlink=True, archive_file=None,
                 lfagmap=False, masterodb=False):
        cprepfile = get_surfex_io_object(cprepfile, filetype="surf", geo=geo, fileformat=csurf_filetype,
                                         lfagmap=lfagmap, masterodb=masterodb,)

        SurfexSurfIO.__init__(self, cprepfile, csurf_filetype, input_file=input_file,
                              archive_file=archive_file, symlink=symlink)
        self.need_pgd = True


class SURFFile(SurfexSurfIO):
    def __init__(self, csurf_filetype, csurffile, geo, archive_file=None, input_file=None, lfagmap=False,
                 masterodb=False):
        csurffile = get_surfex_io_object(csurffile, filetype="surf", geo=geo, fileformat=csurf_filetype,
                                         lfagmap=lfagmap, masterodb=masterodb)

        SurfexSurfIO.__init__(self, csurffile, csurf_filetype, input_file=input_file, archive_file=archive_file)
        self.need_pgd = True


class SurfexFileVariable(object):
    """
    Surfex Variable
    """

    def __init__(self, varname, validtime=None, patches=1, layers=1, basetime=None, interval=None, datatype="float"):
        self.varname = varname
        self.type = datatype
        self.patches = patches
        self.layers = layers
        self.basetime = basetime
        self.interval = interval
        self.validtime = validtime

    def print_var(self):
        return self.varname


def get_surfex_io_object(fname, filetype="surf", fileformat=None, geo=None, lfagmap=False, masterodb=False):

    if filetype is not None:
        if filetype.lower() != "surf" and filetype.lower() != "ts" and filetype.lower() != "forcing":
            raise Exception("Invalid filetype: " + filetype + " Allowed: surf/ts/forcing")

    if fileformat is None:
        fileformat, filetype = guess_file_format(fname, filetype)

    if fileformat.lower() == "ascii":
        if filetype.lower() == "surf":
            obj = AsciiSurfexFile(fname, geo=geo)
        elif filetype.lower() == "forcing":
            raise NotImplementedError("Not implemented yet")
        else:
            raise NotImplementedError

    elif fileformat.lower() == "nc":
        if filetype.lower() == "surf":
            obj = NCSurfexFile(fname, geo=geo)
        else:
            raise NotImplementedError
    elif fileformat.lower() == "netcdf":
        if filetype.lower() == "ts":
            if geo is None:
                raise Exception("Format NetCDF needs a geometry")
            obj = NetCDFSurfexFile(fname, geo)
        elif filetype.lower() == "forcing":
            if geo is None:
                raise Exception("Format NetCDF needs a geometry for reading forcing files")
            obj = ForcingFileNetCDF(fname, geo)
        else:
            raise NotImplementedError
    elif fileformat.lower() == "texte":
        if geo is None:
            raise Exception("Format TEXTE needs a geometry")
        obj = TexteSurfexFile(fname, geo)
    elif fileformat.lower() == "fa":
        if filetype.lower() == "surf":
            obj = FaSurfexFile(fname, geo=geo, lfagmap=lfagmap, masterodb=masterodb)
        else:
            raise NotImplementedError
    # elif fileformat.lower() == "sfx":
    #    if filetype.lower() == "surf":
    #        obj = FaSurfexFile(fname, geo=geo, lfagmap=True, masterodb=masterodb)
    #    else:
    #        raise NotImplementedError
    else:
        raise NotImplementedError("Format not implemented: " + fileformat)

    return obj


def guess_file_format(fname, ftype=None):

    fn = str(os.path.basename(fname))
    ext = str(fn.split(".")[-1])
    if ftype is None:
        surfex.util.info("Trying to guess the file type")
        needles = ["PREP", "PGD"]
        for needle in range(0, len(needles)):
            if re.search(needles[needle], fn):
                ftype = "surf"

        if ext.endswith("nc"):
            ftype = "surf"
        needles = [".*PROGNOSTIC.*", ".*DIAGNOSTICS.*", "SURF_ATM.*"]
        for needle in range(0, len(needles)):
            if re.search(needles[needle], fn):
                ftype = "ts"
        for needle in range(0, len(needles)):
            if ext.endswith("TXT"):
                ftype = "ts"
        needles = ["Forc_.*", "FORCING.*"]
        for needle in range(0, len(needles)):
            if re.search(needles[needle], fn):
                ftype = "forcing"

        if re.search("SURFOUT.*", fn) and ftype is None:
            raise Exception("Can not-auto decide filetype for files called SURFOUT.*.txt. " +
                            "Specify either surf or ts")

    fileformat = None
    surfex.util.info("Trying to guess the file format from extension: " + ext)
    if ext.endswith("txt"):
        fileformat = "ascii"
    if ext.endswith("TXT"):
        fileformat = "texte"
    if ext.endswith("nc"):
        if ftype == "surf":
            fileformat = "nc"
        else:
            fileformat = "netcdf"
    if ext.endswith("fa"):
        fileformat = "fa"
    if ext.endswith("sfx"):
        fileformat = "fa"

    if ftype is None or fileformat is None:
        raise Exception("Filetype and/or format not set: " + str(ftype) + " & " + str(fileformat))
    surfex.util.info("Filetype: " + ftype + " format: " + fileformat)

    return fileformat, ftype


class AsciiSurfexFile(SurfexIO):

    def __init__(self, filename, **kwargs):

        suffix = SurfFileTypeExtension("ASCII", **kwargs).suffix
        self.filename = filename

        geo = None
        if "geo" in kwargs:
            geo = kwargs["geo"]
        if geo is None:
            geo = self.get_geo()

        if not filename.endswith(suffix):
            filename = filename + suffix

        SurfexIO.__init__(self, filename, geo, "txt")

    def get_geo(self):
        if not os.path.isfile(self.filename):
            raise FileNotFoundError("File does not exist: " + str(self.filename))

        grid = self.read("GRID_TYPE", "FULL", "string")
        if len(grid) == 0:
            raise Exception("No grid found")

        if grid[0] == "IGN":
            domain = {
                "nam_ign": {
                    "clambert": self.read("LAMBERT", "&FULL", "integer")[0],
                    "xx": self.read("XX", "&FULL", "float"),
                    "xy": self.read("XY", "&FULL", "float"),
                    "xdx": self.read("XDX", "&FULL", "float"),
                    "xdy": self.read("XY", "&FULL", "float")
                }
            }
            return surfex.geo.IGN(domain)

        elif grid[0] == "LONLATVAL":
            domain = {
                "nam_lonlatval": {
                    "xx": self.read("XX", "&FULL", "float"),
                    "xy": self.read("XY", "&FULL", "float"),
                    "xdx": self.read("DX", "&FULL", "float"),
                    "xdy": self.read("DY", "&FULL", "float")
                }
            }
            return surfex.geo.LonLatVal(domain)

        elif grid[0] == "LONLAT REG":
            domain = {
                "nam_lonlatval_reg": {
                    "lonmin": self.read("LONMIN", "&FULL", "float")[0],
                    "latmin": self.read("LATMIN", "&FULL", "float")[0],
                    "lonmax": self.read("LONMAX", "&FULL", "float")[0],
                    "latmax": self.read("LATMAX", "&FULL", "float")[0],
                    "nlon": self.read("NLON", "&FULL", "integer")[0],
                    "nlat": self.read("NLAT", "&FULL", "integer")[0],
                    "reg_lon": self.read("REG_LON", "&FULL", "float")[0],
                    "reg_lat": self.read("REG_LAT", "&FULL", "float")[0]
                }
            }
            return surfex.geo.LonLatReg(domain)

        elif grid[0] == "CONF PROJ":
            lon0 = self.read("LON0", "&FULL", "float")[0]
            lat0 = self.read("LAT0", "&FULL", "float")[0]
            nx = self.read("IMAX", "&FULL", "integer")[0]
            ny = self.read("JMAX", "&FULL", "integer")[0]
            dx = self.read("XX", "&FULL", "float")[0]
            dy = self.read("XX", "&FULL", "float")[0]

            ll_lon = self.read("LONORI", "&FULL", "float")[0]
            ll_lat = self.read("LATORI", "&FULL", "float")[0]

            earth = 6.37122e+6
            proj_string = "+proj=lcc +lat_0=" + str(lat0) + " +lon_0=" + str(lon0) + " +lat_1=" + \
                          str(lat0) + " +lat_2=" + str(lat0) + " +units=m +no_defs +R=" + str(earth)

            proj = pyproj.CRS.from_string(proj_string)
            wgs84 = pyproj.CRS.from_string("EPSG:4326")
            x0, y0 = pyproj.Transformer.from_crs(wgs84, proj, always_xy=True).transform(ll_lon, ll_lat)
            xc = x0 + 0.5 * (nx - 1) * dx
            yc = y0 + 0.5 * (ny - 1) * dy
            lonc, latc = pyproj.Transformer.from_crs(proj, wgs84, always_xy=True).transform(xc, yc)

            domain = {
                "nam_conf_proj": {
                    "xlon0": lon0,
                    "xlat0": lat0
                },
                "nam_conf_proj_grid": {
                    "xloncen": lonc,
                    "xlatcen": latc,
                    "nimax": nx,
                    "njmax": ny,
                    "xdx": dx,
                    "xdy": dx,
                    "ilone": 0,
                    "ilate": 0
                }
            }
            print(domain)
            return surfex.geo.ConfProj(domain)
        else:
            raise NotImplementedError("Grid " + str(grid[0]) + " not implemented!")

    def read(self, read_par, read_tile, datatype):

        # Add & if not given
        if read_tile.find('&') < 0:
            read_tile = '&' + read_tile
        # print read_tile,read_par
        file = open(self.filename, mode="r")
        read_desc = False
        read_value = False
        values = []
        for line in file:
            # for line in file.read().splitlines():

            # print "T:"+line
            words = line.split()
            if len(words) > 0:
                # print "Line:",read_desc,read_value,":",line
                if read_value and not read_desc:
                    if words[0].find('&') < 0:
                        # print "Value:", line
                        try:
                            if datatype.lower() == "float":
                                for i in range(0, len(words)):
                                    val = float(words[i].replace("D", "E"))
                                    if val == 1e+20:
                                        val = np.nan
                                    values.append(val)
                            elif datatype.lower() == "string":
                                str_words = []
                                for i in range(0, len(words)):
                                    str_words.append(words[i])
                                values.append(" ".join(str_words))
                            elif datatype.lower() == "integer" or datatype.lower() == "int":
                                for i in range(0, len(words)):
                                    values.append(int(words[i]))
                            else:
                                raise NotImplementedError("Type not implemented " + str(datatype))
                        except ValueError:
                            raise Exception('Conversion from ' + str(words) + " to " + str(datatype) +
                                            " does not work! Try a different datatype!")

                if read_desc:
                    # print "Description: ", words[0]
                    read_desc = False
                    read_value = True

                if words[0].find('&') >= 0:
                    tile = words[0]
                    par = words[1]

                    read_value = False
                    if tile.strip().lower() == read_tile.lower() and par.lower() == read_par.lower():
                        read_desc = True
                        read_value = False
                        print("Found:" + str(tile) + " " + str(par))

            # Description could be empty
            else:
                if read_desc:
                    # print "Description: ", words[0]
                    read_desc = False
                    read_value = True

        if len(values) == 0:
            print("No values found!")

        values = np.asarray(values)
        return values

    def field(self, var, validtime=None):

        # TODO
        read_par = var.varname
        read_tile = "&FULL"
        datatype = "float"
        field = self.read(read_par, read_tile, datatype)
        geo_in = self.get_geo()
        field = np.reshape(field, [geo_in.nlons, geo_in.nlats], order="F")
        # field = np.transpose(field)
        return field, geo_in

    def points(self, var, geo_out, validtime=None, interpolation="nearest", cache=None):
        if validtime is not None and type(validtime) != datetime:
            raise Exception("validime must be a datetime object")
        field, geo_in = self.field(var, validtime=validtime)

        points, interpolator = SurfexIO.interpolate_field(self, field, geo_in, geo_out, interpolation=interpolation,
                                                          cache=cache)
        return points, interpolator


class NCSurfexFile(SurfexIO):

    def __init__(self, filename, **kwargs):
        suffix = SurfFileTypeExtension("NC", **kwargs).suffix

        if not filename.endswith(suffix):
            filename = filename + suffix

        geo = None
        if "geo" in kwargs:
            geo = kwargs["geo"]

        if geo is None:
            self.filename = filename
            geo = self.get_geo()

        SurfexIO.__init__(self, filename, geo, "nc")

    def get_geo(self):

        fh = netCDF4.Dataset(self.filename, "r")
        cgrid = str(netCDF4.chartostring(fh["GRID_TYPE"][:])).strip()
        # print(":" + cgrid + ":")
        if cgrid == "CONF PROJ":
            lon0 = fh["LON0"][:]
            lat0 = fh["LAT0"][:]
            nx = int(fh["IMAX"][0])
            ny = int(fh["JMAX"][0])
            dx = float(fh["DX"][0][0])
            dy = float(fh["DY"][0][0])

            ll_lon = fh["LONORI"][:]
            ll_lat = fh["LATORI"][:]
            earth = 6.37122e+6
            proj_string = "+proj=lcc +lat_0=" + str(lat0) + " +lon_0=" + str(lon0) + " +lat_1=" + \
                          str(lat0) + " +lat_2=" + str(lat0) + " +units=m +no_defs +R=" + str(earth)

            proj = pyproj.CRS.from_string(proj_string)
            wgs84 = pyproj.CRS.from_string("EPSG:4326")
            x0, y0 = pyproj.Transformer.from_crs(wgs84, proj, always_xy=True).transform(ll_lon, ll_lat)
            xc = x0 + 0.5 * (nx - 1) * dx
            yc = y0 + 0.5 * (ny - 1) * dy
            lonc, latc = pyproj.Transformer.from_crs(proj, wgs84, always_xy=True).transform(xc, yc)

            domain = {
                "nam_conf_proj": {
                    "xlon0": lon0,
                    "xlat0": lat0  # ,
                    # "xrpk": fh["RPK"][0],
                    # "beta": fh["BETA"][0]
                },
                "nam_conf_proj_grid": {
                    "xloncen": lonc,
                    "xlatcen": latc,
                    "nimax": nx,
                    "njmax": ny,
                    "xdx": dx,
                    "xdy": dy,
                    "ilone": 0,
                    "ilate": 0
                }
            }
            return surfex.geo.ConfProj(domain)
        elif cgrid == "IGN":
            domain = {
                "nam_ign": {
                    "clambert": fh["CLAMBERT"][0],
                    "xx": fh["XX"][:],
                    "xy": fh["XY"][:],
                    "xdx": fh["DX"][:],
                    "xdy": fh["DY"][:]
                }
            }
            return surfex.geo.IGN(domain)

        elif cgrid == "LONLATVAL":
            domain = {
                "nam_lonlatval": {
                    "xx": fh["XX"][:],
                    "xy": fh["XY"][:],
                    "xdx": fh["DX"][:],
                    "xdy": fh["DY"][:]
                }
            }
            return surfex.geo.LonLatVal(domain)

        elif cgrid == "LONLAT REG":
            domain = {
                "nam_lonlatval_reg": {
                    "lonmin": fh["LONMIN"][0],
                    "latmin": fh["LATMIN"][0],
                    "lonmax": fh["LONMAX"][0],
                    "latmax": fh["LATMAX"][0],
                    "nlon": fh["NLON"][0],
                    "nlat": fh["NLAT"][0],
                    "reg_lon": fh["REG_LON"][0],
                    "reg_lat": fh["REG_LAT"][0],
                }
            }
            return surfex.geo.LonLatReg(domain)
        else:
            raise NotImplementedError(cgrid + " is not implemented")

    def field(self, var, validtime=None):

        fh = netCDF4.Dataset(self.filename, "r")
        if validtime is None:
            pass
        elif type(validtime) != datetime:
            raise Exception("validime must be a datetime object")
        else:
            if hasattr(fh, "DTCUR-YEAR"):
                year = fh["DTCUR-YEAR"][0]
                month = fh["DTCUR-MONTH"][0]
                day = fh["DTCUR-DAY"][0]
                time = int(fh["DTCUR-TIME"][0])
                hour = int(time/3600)

                # TODO minutes
                time_in_file = datetime(year=year, month=month, day=day, hour=hour)
                if validtime != time_in_file:
                    print(time_in_file, validtime)
                    raise Exception("Mismatch in times in file and the wanted time")
            else:
                print("Not checking time")

        geo_in = self.get_geo()
        field = fh[var.varname][:]

        # Reshape to fortran 2D style
        # field = np.reshape(field, [geo_in.nlons, geo_in.nlats], order="F")
        # Does not work wih interpolator
        field = np.transpose(field)
        return field, geo_in

    def points(self, var, geo_out, validtime=None, interpolation="nearest", cache=None):

        if validtime is not None and type(validtime) != datetime:
            raise Exception("validime must be a datetime object")
        field, geo_in = self.field(var, validtime=validtime)

        points, interpolator = SurfexIO.interpolate_field(self, field, geo_in, geo_out, interpolation=interpolation,
                                                          cache=cache)
        return points, interpolator


class FaSurfexFile(SurfexIO):

    def __init__(self, filename, **kwargs):

        geo = None
        if "geo" in kwargs:
            geo = kwargs["geo"]
        lfagmap = True
        if "lfagmap" in kwargs:
            lfagmap = kwargs["lfagmap"]

        extension = SurfFileTypeExtension("FA", **kwargs)
        extension_suffix = extension.suffix

        # Surfex binaries use .fa suffix for all input files no matter if they are lfagmap files

        if not filename.endswith(extension_suffix):
            filename = filename + extension_suffix

        # if geo is None:
        #    geo = self.get_geo()

        SurfexIO.__init__(self, filename, geo, extension)
        self.lfagmap = lfagmap

    # def get_geo(self):
    #    # TODO read geo from SURFEX FA file
    #    # geo = None
    #    return None

    def field(self, var, validtime=None):

        fh = surfex.fa.Fa(self.filename)
        if validtime is None:
            pass
        elif type(validtime) != datetime:
            raise Exception("validime must be a datetime object")

        geo_in = self.geo
        field = fh.field(var.varname, validtime)

        # Reshape to fortran 2D style
        field = np.reshape(field, [geo_in.nlons, geo_in.nlats], order="F")
        field = np.transpose(field)
        return field, geo_in

    def points(self, var, geo_out, validtime=None, interpolation="nearest", cache=None):

        if validtime is not None and type(validtime) != datetime:
            raise Exception("validime must be a datetime object")
        field, geo_in = self.field(var, validtime=validtime)

        points, interpolator = SurfexIO.interpolate_field(self, field, geo_in, geo_out, interpolation=interpolation,
                                                          cache=cache)
        return points, interpolator


class SurfFileTypeExtension(object):

    def __init__(self, csurf_filetype, **kwargs):
        masterodb = False
        if "masterodb" in kwargs:
            masterodb = kwargs["masterodb"]
        lfagmap = True
        if "lfagmap" in kwargs:
            lfagmap = kwargs["lfagmap"]

        suffix = None
        extension = ""
        if csurf_filetype.lower() == "nc":
            extension = "nc"
        elif csurf_filetype.lower() == "ascii":
            extension = "txt"
        elif csurf_filetype.lower() == "fa":
            if lfagmap:
                extension = "sfx"
                if not masterodb:
                    suffix = ".fa"
            else:
                extension = "fa"

        if suffix is None:
            suffix = "." + extension

        self.type = extension
        self.suffix = suffix


class NetCDFSurfexFile(SurfexIO):
    """
    Reading surfex NetCDF output
    """

    def __init__(self, filename, geo):
        self.fh = netCDF4.Dataset(filename, "r")
        SurfexIO.__init__(self, filename, geo, "nc")

    def read(self, var, times):
        """

        Read a field, return a 2D array
        :param var:
        :param times:
        :return: values, geo
        """

        layers = var.layers
        patches = var.patches

        if not isinstance(times, (list, tuple)):
            raise Exception("times must be list or tuple")
        if not isinstance(layers, (list, tuple)):
            raise Exception("patches must be list or tuple")
        if not isinstance(patches, (list, tuple)):
            raise Exception("patches must be list or tuple")

        values = np.array([])
        times_read = []
        ndims = 0
        dim_indices = []
        mapping = {}
        npatch = 1

        if self.fh.variables[var].shape[0] > 0:
            # p rint self.fh.variables[var]
            for dim in self.fh.variables[var].dimensions:
                # print dim,ndims
                dimlen = self.fh.variables[var].shape[ndims]
                this_dim = []

                if dim == "time":
                    mapping[0] = ndims
                    times_for_var = self.fh.variables['time']
                    units = times_for_var.units
                    try:
                        t_cal = times_for_var.calendar
                    except AttributeError:  # Attribute doesn't exist
                        t_cal = u"gregorian"  # or standard

                    indices = []
                    [indices.append(i) for i in range(0, dimlen)]
                    times_for_var = netCDF4.num2date(times_for_var[indices], units=units, calendar=t_cal)
                    if len(times) > 0:
                        for t_to_find in range(0, len(times)):
                            for t in range(0, len(indices)):
                                if times_for_var[t] == times[t_to_find]:
                                    # print t, t_to_find, times_for_var[t], times[t_to_find]
                                    this_dim.append(t)
                                    times_read.append(times[t_to_find])
                    else:
                        times_read = times_for_var
                        [this_dim.append(i) for i in range(0, dimlen)]

                elif dim == "Number_of_points":
                    mapping[1] = ndims
                    [this_dim.append(i) for i in range(0, dimlen)]
                elif dim == "xx":
                    mapping[1] = ndims
                    [this_dim.append(i) for i in range(0, dimlen)]
                elif dim == "yy":
                    mapping[2] = ndims
                    [this_dim.append(i) for i in range(0, dimlen)]
                elif dim == "Number_of_Tile":
                    mapping[3] = ndims
                    npatch = dimlen
                    if len(patches) > 0:
                        npatch = len(patches)
                        this_dim = patches
                    else:
                        [this_dim.append(i) for i in range(0, dimlen)]
                elif dim == "Number_of_Layers":
                    mapping[4] = ndims
                    npatch = dimlen
                    if len(layers) > 0:
                        # nlayers = len(layers)
                        this_dim = layers
                    else:
                        [this_dim.append(i) for i in range(0, dimlen)]
                elif dim == "lon":
                    mapping[1] = ndims
                    [this_dim.append(i) for i in range(0, dimlen)]
                elif dim == "lat":
                    mapping[2] = ndims
                    [this_dim.append(i) for i in range(0, dimlen)]
                else:
                    raise NotImplementedError("Not implemented for: " + dim)

                dim_indices.append(this_dim)
                ndims = ndims + 1

            field = self.fh.variables[var][dim_indices]

            # Add extra dimensions
            # print mapping
            i = 0
            reverse_mapping = []
            for d in range(0, 5):
                if d not in mapping:
                    # print "Adding dimension " + str(d)
                    field = np.expand_dims(field, len(dim_indices) + i)
                    reverse_mapping.append(len(dim_indices) + i)
                    i = i + 1
                else:
                    reverse_mapping.append(mapping[d])

            # Transpose to 5D array
            # print "Transpose to 5D array"
            # print reverse_mapping
            field = np.transpose(field, reverse_mapping)
            npoints = self.geo.npoints * npatch

            # Create 2-D array with times and points as for the other formats
            for t in range(0, field.shape[0]):
                field2d = np.empty(npoints)
                i = 0
                # print t,npatch,npoints,field.shape,field2d.shape
                for p in range(0, npatch):
                    if self.geo.mask is not None:
                        ii = 0
                        j = 0
                        # For some reason the NetCDF files have one less dimension in all
                        # dimensions than the PGD dimension and mask needs x first.
                        for x in range(-1, field.shape[1] + 1):
                            for y in range(-1, field.shape[2] + 1):
                                if x in range(0, field.shape[1]) and y in range(0, field.shape[2]):
                                    # print i, ii,j, t, x, y, p, self.geo.mask[j]
                                    if self.geo.mask[j] == ii:
                                        field2d[i] = np.nan
                                        if field[t, x, y, p] != np.nan:
                                            field2d[i] = field[t, x, y, p]
                                        i = i + 1
                                        j = j + 1
                                ii = ii + 1
                    else:
                        for y in range(0, field.shape[2]):
                            for x in range(0, field.shape[1]):
                                field2d[i] = np.nan
                                if field[t, x, y, p] != np.nan:
                                    field2d[i] = field[t, x, y, p]
                                i = i + 1
                if i != npoints:
                    raise Exception("Mismatch in points " + str(i) + "!=" + str(npoints))

                values = np.append(values, field2d)
            # Re-shape to proper format
            values = np.reshape(values, [field.shape[0], npoints])

        else:
            raise Exception("Variable " + var + " not found!")

        return values, self.geo

    def field(self, var, validtime=None):

        if validtime is None:
            validtime = []
            # raise Exception("You must set times to read forcing data")
        elif type(validtime) != datetime:
            raise Exception("validime must be a datetime object")
        else:
            validtime = [validtime]

        field, geo_in = self.read(var, validtime)
        # Reshape to fortran 2D style
        field = np.reshape(field, [geo_in.nlons, geo_in.nlats], order="F")
        return field, geo_in

    def points(self, var, geo_out, validtime=None, interpolation="nearest", cache=None):
        if validtime is not None and type(validtime) != datetime:
            raise Exception("validime must be a datetime object")
        field, geo_in = self.field(var, validtime=validtime)

        points, interpolator = SurfexIO.interpolate_field(self, field, geo_in, geo_out, interpolation=interpolation,
                                                          cache=cache)
        return points, interpolator


class TexteSurfexFile(SurfexIO):

    """
    Reading surfex TEXTE output
    """

    def __init__(self, filename, geo):
        self.file = None
        SurfexIO.__init__(self,  filename, geo, "TXT")

    def read(self, variable, times):
        self.file = open(self.filename, mode="r")

        base_time = variable.basetime
        interval = variable.interval
        npatch = variable.patches

        if base_time is None:
            raise Exception("Basetime must be set for TEXTE")
        if interval is None:
            raise Exception("Interval must be set for TEXTE")

        if not isinstance(times, (list, tuple)):
            raise Exception("times must be list or tuple")

        values = np.array([])
        times_read = np.array([])
        end_of_line = self.geo.npoints * npatch
        this_time = np.empty(self.geo.npoints * npatch)

        t = 1
        col = 0
        for line in self.file.read().splitlines():

            words = line.split()
            if len(words) > 0:
                for i in range(0, len(words)):
                    val = float(words[i].replace("D", "E"))
                    if val == 1e+20:
                        val = np.nan
                    this_time[col] = val

                    col = col + 1
                    if col == end_of_line:

                        if times is None or (base_time + timedelta(seconds=(t * interval))) in times:
                            values = np.append(values, this_time)
                            times_read = np.append(times_read, base_time + timedelta(seconds=(t * interval)))
                            # print i, col, base_time + timedelta(seconds=(t * interval)), this_time

                        t = t + 1
                        col = 0
                        this_time[:] = np.nan
                        if i != len(words) - 1:
                            raise Exception("Dimension of domain does not match end of line! " +
                                            str(i) + " != " + str(len(words) - 1))

        if times_read.shape[0] > 0:
            values = np.reshape(values, [times_read.shape[0], this_time.shape[0]])
        else:
            print("No data found!")

        # print values.shape
        self.file.close()
        return values, self.geo

    def field(self, var, validtime=None):
        if validtime is None:
            raise Exception("You must set times to read forcing data")
        elif type(validtime) != datetime:
            raise Exception("validime must be a datetime object")
        else:
            validtime = [validtime]

        field, geo_in = self.read(var, validtime)
        # Reshape to fortran 2D style
        field = np.reshape(field, [geo_in.nlons, geo_in.nlats], order="F")
        return field, geo_in

    def points(self, var, geo_out, validtime=None, interpolation="nearest", cache=None):

        if validtime is not None and type(validtime) != datetime:
            raise Exception("validime must be a datetime object")
        field, geo_in = self.field(var, validtime=validtime)

        points, interpolator = SurfexIO.interpolate_field(self, field, geo_in, geo_out, interpolation=interpolation,
                                                          cache=cache)
        return points, interpolator


class ForcingFileNetCDF(SurfexIO):

    def __init__(self, fname, geo):
        self.fname = fname
        self.fh = netCDF4.Dataset(fname, "r")
        self.lons = self.fh.variables["LON"]
        self.lats = self.fh.variables["LAT"]
        self.nx = self.lons.shape[0]
        self.ny = self.lats.shape[0]
        SurfexIO.__init__(self, fname, geo, "nc")

    def read_field(self, variable, times):

        var = variable.varname
        field = None
        if self.fh.variables[var].shape[0] > 0:
            if len(self.fh.variables[var].dimensions) == 1:
                dimlen = self.fh.variables[var].shape[0]
                field = self.fh.variables[var][0:dimlen]
            else:
                if len(times) == 0:
                    raise Exception("You must set time!")

                times_read = []
                ndims = 0
                npoints = 0
                for dim in self.fh.variables[var].dimensions:
                    dimlen = self.fh.variables[var].shape[ndims]

                    if dim == "time":
                        times_for_var = self.fh.variables['time']
                        units = times_for_var.units
                        try:
                            t_cal = times_for_var.calendar
                        except AttributeError:  # Attribute doesn't exist
                            t_cal = u"gregorian"  # or standard

                        indices = []
                        [indices.append(i) for i in range(0, dimlen)]
                        times_for_var = netCDF4.num2date(times_for_var[indices], units=units, calendar=t_cal)
                        # print(times_for_var)
                        for times_to_read in range(0, len(times)):
                            # print(times_to_read, times[times_to_read])
                            for t in range(0, len(times_for_var)):
                                # print(t, times_for_var[t], times[times_to_read])
                                test_time = times_for_var[t].strftime("%Y%m%d%H")
                                test_time = datetime.strptime(test_time, "%Y%m%d%H")
                                if test_time == times[times_to_read]:
                                    times_read.append(t)
                                    print(t, times[times_to_read])
                    else:
                        npoints = dimlen

                    ndims = ndims + 1

                if npoints == 0:
                    raise Exception("No points found")

                if len(times_read) == 0 and len(times) > 0:
                    print(times)
                    raise Exception("Valid time not found in file!")

                field = self.fh.variables[var][times_read, 0: npoints]

        return field, self.geo

    def field(self, var, validtime=None):
        if validtime is None:
            validtime = []
            # raise Exception("You must set times to read forcing data")
        elif type(validtime) != datetime:
            raise Exception("validime must be a datetime object")
        else:
            validtime = [validtime]

        field, geo_in = self.read_field(var, validtime)
        # Reshape to fortran 2D style
        field = np.reshape(field, [geo_in.nlons, geo_in.nlats], order="F")
        return field, geo_in

    def points(self, var, geo_out, validtime=None, interpolation=None, cache=None):

        if validtime is not None and type(validtime) != datetime:
            raise Exception("validime must be a datetime object")
        field, geo_in = self.field(var, validtime=validtime)

        points, interpolator = SurfexIO.interpolate_field(self, field, geo_in, geo_out, interpolation=interpolation,
                                                          cache=cache)
        return points, interpolator


def read_surfex_field(varname, filename, validtime=None, basetime=None, patches=-1, layers=-1, fileformat=None,
                      filetype=None, geo=None, datatype=None, interval=None):

    if fileformat is None:
        fileformat, filetype = surfex.file.guess_file_format(filename, filetype)

    if filetype == "surf":
        if fileformat.lower() == "ascii":
            geo = surfex.file.AsciiSurfexFile(filename).geo
        elif fileformat.lower() == "nc":
            geo = surfex.file.NCSurfexFile(filename).geo
        else:
            if geo is None:
                raise NotImplementedError("Not implemnted and geo is None")
    elif geo is None:
        raise Exception("You need to provide a geo object. Filetype is: " + str(filetype))

    sfx_io = surfex.file.get_surfex_io_object(filename, filetype=filetype, fileformat=fileformat, geo=geo)
    var = surfex.file.SurfexFileVariable(varname, validtime=validtime, patches=patches, layers=layers,
                                         basetime=basetime, interval=interval, datatype=datatype)
    field, geo_out = sfx_io.field(var, validtime=validtime)
    return field


def read_surfex_points(varname, filename, geo_out, validtime=None, basetime=None, patches=-1, layers=-1,
                       fileformat=None, filetype=None, geo=None, datatype=None, interval=None, interpolation="nearest"):

    if fileformat is None:
        fileformat, filetype = surfex.file.guess_file_format(filename, filetype)

    if filetype == "surf":
        if fileformat.lower() == "ascii":
            geo = surfex.file.AsciiSurfexFile(filename).geo
        elif fileformat.lower() == "nc":
            geo = surfex.file.NCSurfexFile(filename).geo
        else:
            if geo is None:
                raise NotImplementedError("Not implemnted and geo is None")
    elif geo is None:
        raise Exception("You need to provide a geo object. Filetype is: " + str(filetype))

    sfx_io = surfex.file.get_surfex_io_object(filename, filetype=filetype, fileformat=fileformat, geo=geo)
    var = surfex.file.SurfexFileVariable(varname, validtime=validtime, patches=patches, layers=layers,
                                         basetime=basetime, interval=interval, datatype=datatype)
    field, geo_out = sfx_io.points(var, geo_out, validtime=validtime, interpolation=interpolation)
    return field


def parse_filepattern(file_pattern, basetime, validtime):

    if basetime is None or validtime is None:
        return file_pattern

    # print(file_pattern)
    file_name = str(file_pattern)
    year = basetime.strftime('%Y')
    year2 = basetime.strftime('%y')
    month = basetime.strftime('%m')
    day = basetime.strftime('%d')
    hour = basetime.strftime('%H')
    mins = basetime.strftime('%M')
    dt = validtime-basetime
    l1 = "%d" % (dt.seconds / 3600)
    ll = "%02d" % (dt.seconds / 3600)
    lll = "%03d" % (dt.seconds / 3600)
    llll = "%04d" % (dt.seconds / 3600)
    file_name = file_name.replace('@YYYY@', year)
    file_name = file_name.replace('@YY@', year2)
    file_name = file_name.replace('@MM@', month)
    file_name = file_name.replace('@DD@', day)
    file_name = file_name.replace('@HH@', hour)
    file_name = file_name.replace('@mm@', mins)
    file_name = file_name.replace('@L@', l1)
    file_name = file_name.replace('@LL@', ll)
    file_name = file_name.replace('@LLL@', lll)
    file_name = file_name.replace('@LLLL@', llll)
    return file_name



import unittest
import surfex


class Oi2SodaTest(unittest.TestCase):

    def setUp(self):
        self.testdata = "testdata/"

    def test_oi2soda(self):

        argv = [
            "--t2m_file", self.testdata + "/unittest_an_t2m.nc",
            "--t2m_var", "air_temperature_2m",
            "--rh2m_file", self.testdata + "/unittest_an_rh2m.nc",
            "--rh2m_var", "relative_humidity_2m",
            "--sd_file", self.testdata + "/unittest_an_sd.nc",
            "--sd_var", "surface_snow_thickness",
            "--debug",
            "-o", "unittest_OBSERVATIONS_200330H06.DAT",
            "2020033006"
            ]
        kwargs = surfex.parse_args_oi2soda(argv)
        surfex.run_oi2soda(**kwargs)

    def test_oi2soda_only_rh(self):
        argv = [
            "--rh2m_file", self.testdata + "/unittest_an_rh2m.nc",
            "--rh2m_var", "relative_humidity_2m",
            "--debug",
            "-o", "unittest_OBSERVATIONS_200330H06.DAT",
            "2020033006"
            ]
        kwargs = surfex.parse_args_oi2soda(argv)
        surfex.run_oi2soda(**kwargs)



import unittest
import surfex
import json


class GeoTest(unittest.TestCase):

    # def test_geo_base_not_implemented(self):
    #    my_settings = {}
    #    with self.assertRaises(NotImplementedError):
    #        my_geo = surfex.geo.SurfexGeo(None, 1, 1, 1, 1, 1)
    #        my_geo.update_namelist(my_settings)

    def setUp(self):
        self.domain_conf_proj = {
            "nam_conf_proj_grid": {
                "xlatcen": 60,
                "ilone": 1,
                "xdx": 2500.0,
                "njmax": 2,
                "xloncen": 10,
                "xdy": 2500.0,
                "nimax": 3,
                "ilate": 2
            },
            "nam_pgd_grid": {
                "cgrid": "CONF PROJ"
            },
            "nam_conf_proj": {
                "xlon0": 0,
                "xlat0": 50,
            }
        }

    def test_geo_not_defined(self):
        domain = {"nam_pgd_grid": {"cgrid": "not_existing"}}
        with self.assertRaises(NotImplementedError):
            surfex.geo.get_geo_object(domain)

    def test_get_geo_obj(self):
        domain = {"not_existing": {"some_key": "some_value"}}
        with self.assertRaises(KeyError):
            surfex.geo.get_geo_object(domain)

        domain = {"nam_pgd_grid": {"not_existing": "some_value"}}
        with self.assertRaises(KeyError):
            surfex.geo.get_geo_object(domain)

    def test_geo_conf_proj(self):

        my_geo = surfex.geo.get_geo_object(self.domain_conf_proj)

        json_settings = {"nam_io_offline": {"csurf_filetype": "NC"}}
        my_settings = surfex.BaseNamelist.ascii2nml(json_settings)
        my_geo.update_namelist(my_settings)
        self.assertEqual(self.domain_conf_proj["nam_pgd_grid"]["cgrid"], my_geo.cgrid)
        print(my_geo.identifier())

        new_domain = {"not_existing": {"not_existing": "some_value"},
                      "nam_conf_proj_grid": self.domain_conf_proj["nam_conf_proj_grid"]}
        with self.assertRaises(KeyError):
            surfex.geo.ConfProj(new_domain)

        new_domain = {"not_existing": {"not_existing": "some_value"},
                      "nam_conf_proj": self.domain_conf_proj["nam_conf_proj"]}
        with self.assertRaises(KeyError):
            surfex.geo.ConfProj(new_domain)

        new_domain = {"nam_conf_proj": {"not_existing": "some_value"},
                      "nam_conf_proj_grid": self.domain_conf_proj["nam_conf_proj_grid"]}
        with self.assertRaises(KeyError):
            surfex.geo.ConfProj(new_domain)

        new_domain = {"nam_conf_proj_grid": {"not_existing": "some_value"},
                      "nam_conf_proj": self.domain_conf_proj["nam_conf_proj"]}
        with self.assertRaises(KeyError):
            surfex.geo.ConfProj(new_domain)

    def test_geo_lonlat_reg(self):
        domain = {
            "nam_pgd_grid": {
                "cgrid": "LONLAT REG"
            },
            "nam_lonlat_reg": {
                "xlonmin": 10,
                "xlonmax": 11,
                "xlatmin": 60,
                "xlatmax": 61,
                "nlon": 11,
                "nlat": 11
            }
        }
        my_geo = surfex.get_geo_object(domain)
        json_settings = {"nam_io_offline": {"csurf_filetype": "NC"}}
        my_settings = surfex.BaseNamelist.ascii2nml(json_settings)
        my_settings = my_geo.update_namelist(my_settings)
        self.assertEqual(domain["nam_pgd_grid"]["cgrid"], my_geo.cgrid)
        self.assertEqual(my_settings["nam_pgd_grid"]["cgrid"], my_geo.cgrid)

        domain = {
            "nam_pgd_grid": {
                "cgrid": "LONLAT REG"
            },
            "nam_lonlat_reg": {
                "xlonmin": 10,
                "xlonmax": 11,
                "xlatmin": 60,
                "xlatmax": 61,
                "nlon": 0,
                "nlat": 11
            }
        }
        with self.assertRaises(ZeroDivisionError):
            surfex.geo.LonLatReg(domain)

        domain = {"not_existing": {"existing": "some_value"}}
        with self.assertRaises(KeyError):
            surfex.geo.LonLatReg(domain)

        domain = {"nam_lonlat_reg": {"not_existing": "some_value"}}
        with self.assertRaises(KeyError):
            surfex.geo.LonLatReg(domain)

    def test_geo_lonlatval(self):
        domain = {
            "nam_pgd_grid": {
                "cgrid": "LONLATVAL"
            },
            "nam_lonlatval": {
                "xx": [10.0, 11.0],
                "xy": [60.0, 61.0],
                "xdx": [0.1, 0.1],
                "xdy": [0.1, 0.1]
            }
        }
        my_geo = surfex.get_geo_object(domain)
        json_settings = {"nam_io_offline": {"csurf_filetype": "NC"}}
        my_settings = surfex.BaseNamelist.ascii2nml(json_settings)
        my_settings = my_geo.update_namelist(my_settings)
        self.assertEqual(domain["nam_pgd_grid"]["cgrid"], my_geo.cgrid)
        self.assertEqual(my_settings["nam_pgd_grid"]["cgrid"], my_geo.cgrid)

        domain = {"not_existing": {"existing": "some_value"}}
        with self.assertRaises(KeyError):
            surfex.geo.LonLatVal(domain)

        domain = {"nam_lonlatval": {"not_existing": "some_value"}}
        with self.assertRaises(KeyError):
            surfex.geo.LonLatVal(domain)

    def test_geo_cartesian(self):
        domain = {
            "nam_pgd_grid": {
                "cgrid": "CARTESIAN"
            },
            "nam_cartesian": {
                "xlat0": 0,
                "xlon0": 0,
                "nimax": 11,
                "njmax": 21,
                "xdx": 0.1,
                "xdy": 0.05
            }
        }
        my_geo = surfex.get_geo_object(domain)
        json_settings = {"nam_io_offline": {"csurf_filetype": "NC"}}
        my_settings = surfex.BaseNamelist.ascii2nml(json_settings)
        my_settings = my_geo.update_namelist(my_settings)
        self.assertEqual(domain["nam_pgd_grid"]["cgrid"], my_geo.cgrid)
        self.assertEqual(my_settings["nam_pgd_grid"]["cgrid"], my_geo.cgrid)

        domain = {"not_existing": {"existing": "some_value"}}
        with self.assertRaises(KeyError):
            surfex.geo.Cartesian(domain)

        domain = {"nam_cartesian": {"not_existing": "some_value"}}
        with self.assertRaises(KeyError):
            surfex.geo.Cartesian(domain)

    def test_geo_ign(self):
        domain = {
            "nam_pgd_grid": {
                "cgrid": "IGN"
            },
            "nam_ign": {
                "clambert": 7,
                "npoints": 3,
                "xx": [11000, 13000, 11000],
                "xy": [21000, 21000, 23000],
                "xdx": [1000, 1000, 1000],
                "xdy": [1000, 1000, 1000],
                "xx_llcorner": 0,
                "xy_llcorner": 0,
                "xcellsize": 1000,
                "ncols": 1,
                "nrows": 1
            }
        }
        my_geo = surfex.geo.IGN(domain, recreate=True)
        json_settings = {"nam_io_offline": {"csurf_filetype": "NC"}}
        my_settings = surfex.BaseNamelist.ascii2nml(json_settings)
        my_settings = my_geo.update_namelist(my_settings)
        self.assertEqual(domain["nam_pgd_grid"]["cgrid"], my_geo.cgrid)
        self.assertEqual(my_settings["nam_pgd_grid"]["cgrid"], my_geo.cgrid)

        my_geo1 = surfex.geo.IGN(domain, recreate=False)
        my_geo2 = surfex.geo.IGN(domain, recreate=True)
        self.assertTrue(my_geo1.is_identical(my_geo2))

        domain = {
            "nam_pgd_grid": {
                "cgrid": "IGN"
            },
            "nam_ign": {
                "clambert": -99,
                "npoints": 0,
                "xx": 11,
                "xy": 21,
                "xdx": 1000,
                "xdy": 1000,
                "xx_llcorner": 0,
                "xy_llcorner": 0,
                "xcellsize": 1000,
                "ncols": 1,
                "nrows": 1
            }
        }
        with self.assertRaises(NotImplementedError):
            surfex.get_geo_object(domain)

        domain = {"not_existing": {"existing": "some_value"}}
        with self.assertRaises(KeyError):
            surfex.geo.IGN(domain)

        domain = {"nam_ign": {"not_existing": "some_value"}}
        with self.assertRaises(KeyError):
            surfex.geo.IGN(domain)

    def test_set_domain(self):
        domains = {"NAME": {"nam_pgd_grid": {"cgrid": "some_projection"}}}
        domain = surfex.geo.set_domain(domains, "NAME")
        self.assertEqual(domains["NAME"]["nam_pgd_grid"]["cgrid"], domain["nam_pgd_grid"]["cgrid"])

        with self.assertRaises(Exception):
            surfex.geo.set_domain(domains, "not_existing")

        domains = ["NAME"]
        with self.assertRaises(Exception):
            surfex.geo.set_domain(domains, "NAME")

        domains = json.load(open("test/settings/domains.json", "r"))
        domain_name = "CONF_PROJ_TEST"
        domain_json = surfex.geo.set_domain(domains, domain_name)
        saved_domain = json.load(open("test/settings/conf_proj_test.json"))
        self.assertDictEqual(domain_json, saved_domain)


if __name__ == '__main__':
    unittest.main()



import os
import surfex
from surfex.util import data_merge
import copy
from abc import abstractmethod, ABCMeta
import numpy as np
try:
    from StringIO import StringIO   # Python 2.x
except ImportError:
    from io import StringIO         # Python 3.x


class ReadData(object):
    __metaclass__ = ABCMeta

    def __init__(self, geo, var_name):
        self.geo = geo
        self.var_name = var_name
        # print "Constructed "+self.__class__.__name__+" for " + self.var_name

    @abstractmethod
    def read_time_step(self, validtime, cache):
        raise NotImplementedError('users must define read_time_step to use this base class')

    @abstractmethod
    def print_info(self):
        raise NotImplementedError('users must define read_time_step to use this base class')


# Direct data can be ead with this class with converter = None
class ConvertedInput(ReadData):
    """
    Converted input

    Args:
        geo (surfex.Geo): Surfex geometry
        var_name (str): Variable name
        converter (surfex.Converter): Converter
    """

    def __init__(self, geo, var_name, converter):
        ReadData.__init__(self, geo, var_name)
        self.converter = converter

    def read_time_step(self, validtime, cache):
        field = self.converter.read_time_step(self.geo, validtime, cache)
        # Preserve positive values for precipitation
        # TODO
        # if self.var_name == "RAIN" or self.var_name == "SNOW":
        #    field[field < 0.] = 0.
        return field

    def print_info(self):
        self.converter.print_info()


class ConstantValue(ReadData):
    """
    Set a field to a constant value

    :param geo: Geometry object
    :type geo: surfex.Geo
    :param var_name: Variable name
    :type var_name: str
    :param var_dict: Variable definition
    :type var_dict: dict
    """
    def __init__(self, geo, var_name, var_dict):
        ReadData.__init__(self, geo, var_name)
        self.var_dict = var_dict
        if "value" in self.var_dict:
            self.value = self.var_dict["value"]
        else:
            print("Constant value must have a value!")
            raise

    def read_time_step(self, validtime, cache):
        field = np.array([float(i) for i in range(0, self.geo.npoints)])
        field.fill(self.value)
        # print field.shape
        return field

    def print_info(self):
        print(self.var_name)


def remove_existing_file(f_in, f_out):
    if f_in is None:
        raise FileNotFoundError("Input file not set")
    # If files are not the same file
    if os.path.abspath(f_in) != os.path.abspath(f_out):
        if os.path.isdir(f_out):
            raise IsADirectoryError(f_out + " is a directory! Please remove it if desired")
        if os.path.islink(f_out):
            os.unlink(f_out)
        if os.path.isfile(f_out):
            os.remove(f_out)
    # files have the same path. Remove if it is a symlink
    else:
        if os.path.islink(f_out):
            os.unlink(f_out)


#######################################################
#######################################################


class Converter:
    """
    Main interface to read a field is done through a converter.
    The converter name is default "None" to read a plain field without any conversion.

    Args:
        name (str): name of the converter
        validtime (datetime.datetime): The valid time you want to read
        defs (dict): A dictionary defining the variables
        conf (dict): A dictionary defining the converter
        fileformat (str): Fileformat of the converter
        basetime (datetime.datetime): The base time of the input data source
        debug (bool): Debug option

    """

    def __init__(self, name, validtime, defs, conf, fileformat, basetime, debug=False):
        """
        Initializing the converter
        """

        self.name = name
        self.validtime = validtime
        self.basetime = basetime
        # self.intervall = intervall

        print(conf, self.name)
        if self.name not in conf:
            print(conf)
            raise KeyError(self.name + " is missing in converter definition")

        if self.name == "none":
            self.var = self.create_variable(fileformat, defs, conf[self.name], debug)
        elif name == "rh2q":
            self.rh = self.create_variable(fileformat, defs, conf[self.name]["rh"], debug)
            self.t = self.create_variable(fileformat, defs, conf[self.name]["t"], debug)
            self.p = self.create_variable(fileformat, defs, conf[self.name]["p"], debug)
        elif name == "windspeed" or name == "winddir":
            self.x = self.create_variable(fileformat, defs, conf[self.name]["x"], debug)
            self.y = self.create_variable(fileformat, defs, conf[self.name]["y"], debug)
        elif name == "totalprec":
            self.totalprec = self.create_variable(fileformat, defs, conf[self.name]["totalprec"], debug)
            self.snow = self.create_variable(fileformat, defs, conf[self.name]["snow"], debug)
        elif name == "calcsnow":
            self.totalprec = self.create_variable(fileformat, defs, conf[self.name]["totalprec"], debug)
            self.t = self.create_variable(fileformat, defs, conf[self.name]["t"], debug)
            # self.rh = self.create_variable(fileformat,defs,conf[self.name]["t"],debug)
            # self.p = self.create_variable(fileformat,defs,conf[self.name]["p"],debug)
        elif name == "calcrain":
            self.totalprec = self.create_variable(fileformat, defs, conf[self.name]["totalprec"], debug)
            self.t = self.create_variable(fileformat, defs, conf[self.name]["t"], debug)
        elif name == "phi2m":
            self.phi = self.create_variable(fileformat, defs, conf[self.name]["phi"], debug)
        elif self.name == "swe2sd":
            self.swe = self.create_variable(fileformat, defs, conf[self.name]["swe"], debug)
        elif self.name == "sweclim":
            self.swe = self.create_variable(fileformat, defs, conf[self.name]["swe"], debug)
        elif self.name == "sea2land":
            self.sea = self.create_variable(fileformat, defs, conf[self.name]["sea"], debug)
        elif self.name == "tap":
            self.tap1 = self.create_variable(fileformat, defs, conf[self.name]["tap1"], debug)
            self.tap2 = self.create_variable(fileformat, defs, conf[self.name]["tap2"], debug)
        elif self.name == "rhp":
            self.rhp1 = self.create_variable(fileformat, defs, conf[self.name]["rhp1"], debug)
            self.rhp2 = self.create_variable(fileformat, defs, conf[self.name]["rhp2"], debug)
        elif self.name == "sdp":
            self.sdp1 = self.create_variable(fileformat, defs, conf[self.name]["sdp1"], debug)
            self.sdp2 = self.create_variable(fileformat, defs, conf[self.name]["sdp2"], debug)
        else:
            print("Converter " + self.name + " not implemented")
            raise NotImplementedError

        # print "Constructed the converter " + self.name

    def print_info(self):
        print(self.name)

    def create_variable(self, fileformat, defs, var_dict, debug):
        """
        Create a variable

        Args:
            fileformat (str): Fileformat
            defs (dict): defs
            var_dict (dict): Variable dictionary
            debug(bool): Debug option

        Raises:
             NotImplementedError: Not implemented

        Returns:
            field: The read field
        """

        # Finally we can merge the variable with the default settings
        # Create deep copies not to inherit between variables
        defs = copy.deepcopy(defs)
        var_dict = copy.deepcopy(var_dict)
        merged_dict = data_merge(defs, var_dict)

        if fileformat == "netcdf":
            var = surfex.variable.NetcdfVariable(merged_dict, self.basetime, self.validtime,  debug=debug)
        elif fileformat == "grib1" or fileformat == "grib2":
            var = surfex.variable.GribVariable(merged_dict, self.basetime, self.validtime,  debug=debug,
                                               grib_type=fileformat)
        elif fileformat == "surfex":
            var = surfex.variable.SurfexVariable(merged_dict, self.basetime, self.validtime, debug=debug)
        elif fileformat == "fa":
            var = surfex.variable.FaVariable(merged_dict, self.basetime, self.validtime, debug=debug)
        elif fileformat == "obs":
            var = surfex.variable.ObservationVariable(merged_dict, self.basetime, self.validtime, debug=debug)
        elif fileformat == "constant":
            raise NotImplementedError("Create variable for format " + fileformat + " not implemented!")
        else:
            raise NotImplementedError("Create variable for format " + fileformat + " not implemented!")

        # TODO: Put this under verbose flag and format printing
        # var.print_variable_info()
        return var

    def read_time_step(self, geo, validtime, cache):
        # print("Time in converter: "+self.name+" "+validtime.strftime('%Y%m%d%H'))

        gravity = 9.81
        field = np.empty(geo.npoints)
        # Specific reading for each converter
        if self.name == "none":
            field = self.var.read_variable(geo, validtime, cache)
        elif self.name == "windspeed" or self.name == "winddir":
            field_x = self.x.read_variable(geo, validtime, cache)
            field_y = self.y.read_variable(geo, validtime, cache)
            # field_y = self.y.read_variable(geo,validtime,cache)
            if self.name == "windspeed":
                field = np.sqrt(np.square(field_x) + np.square(field_y))
                np.where(field < 0.005, field, 0)
            elif self.name == "winddir":
                # TODO chek formulation
                field = np.mod(np.rad2deg(np.arctan2(field_x, field_y)) + 180, 360)

        elif self.name == "rh2q":
            field_rh = self.rh.read_variable(geo, validtime, cache)  # %
            field_t = self.t.read_variable(geo, validtime, cache)  # In K
            field_p = self.p.read_variable(geo, validtime, cache)  # In Pa

            field_p_mb = np.divide(field_p, 100.)
            field_t_c = np.subtract(field_t, 273.15)

            exp = np.divide(np.multiply(17.67, field_t_c), np.add(field_t_c, 243.5))
            es = np.multiply(6.112, np.exp(exp))
            field = np.divide(np.multiply(0.622, field_rh / 100.) * es, field_p_mb)

            # ZES = 6.112 * exp((17.67 * (ZT - 273.15)) / ((ZT - 273.15) + 243.5))
            # ZE = ZRH * ZES
            # ZRATIO = 0.622 * ZE / (ZPRES / 100.)
            # RH2Q = 1. / (1. / ZRATIO + 1.)
        elif self.name == "totalprec":
            field_totalprec = self.totalprec.read_variable(geo, validtime, cache)
            field_snow = self.snow.read_variable(geo, validtime, cache)
            field = np.subtract(field_totalprec, field_snow)
        elif self.name == "calcrain":
            field_totalprec = self.totalprec.read_variable(geo, validtime, cache)
            field_t = self.t.read_variable(geo, validtime, cache)
            field = field_totalprec
            field[field_t < 1] = 0
        elif self.name == "calcsnow":
            field_totalprec = self.totalprec.read_variable(geo, validtime, cache)
            # field_rh = self.rh.read_variable(geo, validtime,cache) #
            field_t = self.t.read_variable(geo, validtime, cache)  # In K
            # field_p = self.p.read_variable(geo, validtime,cache)   # In Pa
            # tc = field_t + 273.15
            # e  = (field_rh)*0.611*exp((17.63*tc)/(tc+243.04));
            # Td = (116.9 + 243.04*log(e))/(16.78-log(e));
            # gamma = 0.00066 * field_p/1000;
            # delta = (4098*e)/pow(Td+243.04,2);
            # if(gamma + delta == 0):
            # print("problem?")
            # wetbulbTemperature = (gamma * tc + delta * Td)/(gamma + delta);
            # wetbulbTemperatureK  = wetbulbTemperature + 273.15;
            field = field_totalprec
            field[field_t > 1] = 0
        elif self.name == "phi2m":
            field = self.phi.read_variable(geo, validtime, cache)
            field = np.divide(field, gravity)
            field[(field < 0)] = 0.
        elif self.name == "swe2sd":
            field = self.swe.read_variable(geo, validtime, cache)
            rho = self.swe.read_variable(geo, validtime, cache)
            field = np.divide(field, rho)
        elif self.name == "sweclim":
            field = self.swe.read_variable(geo, validtime, cache)
            rhoclim = {"01": 222., "02": 233., "03": 240., "04": 278., "05": 212., "06": 312., "07": 312., "08": 143.,
                       "09": 143., "10": 161., "11": 182., "12": 213.}
            month = validtime.strftime("%m")
            if month in rhoclim:
                field = np.divide(field, rhoclim[month])
            else:
                raise Exception("Could not found climatological mean for month " + str(month))
        elif self.name == "sea2land":
            field = self.sea.read_variable(geo, validtime, cache)
            field = np.subtract(1, field)
        elif self.name == "tap":
            tap1 = self.tap1.read_variable(geo, validtime, cache)
            tap2 = self.tap2.read_variable(geo, validtime, cache)
            field = np.where(np.isnan(tap1), tap2, tap1)
        elif self.name == "rhp":
            rhp1 = self.rhp1.read_variable(geo, validtime, cache)
            rhp2 = self.rhp2.read_variable(geo, validtime, cache)
            field = np.where(np.isnan(rhp1), rhp2, rhp1)
        elif self.name == "sdp":
            sdp1 = self.rhp1.read_variable(geo, validtime, cache)
            sdp2 = self.rhp2.read_variable(geo, validtime, cache)
            field = np.where(np.isnan(sdp1), sdp2, sdp1)
        else:
            print("Converter " + self.name + " not implemented")
            raise NotImplementedError
        return field



import unittest
import surfex


class PlotTest(unittest.TestCase):

    def setUp(self):
        self.testdata = "testdata/"

    def test_plot_grib1(self):
        argv = [
            "-it", "grib1",
            "-t", "2020111306",
            "-g", "test/settings/conf_proj_test.json",
            "--indicatorOfParameter", "11",
            "--level", "2",
            "--levelType", "105",
            "-i", self.testdata + "/fc2020111303+0003grib1",
            "-o", "unittest_output_plot_grib1.png",
            "--debug"
            ]
        kwargs = surfex.parse_args_plot_field(argv)
        surfex.run_plot_field(**kwargs)

    def test_plot_grib2(self):
        argv = [
            "-it", "grib2",
            "-t", "2020111306",
            "-g", "test/settings/conf_proj_test.json",
            "--levelType", "103",
            "--discipline", "0",
            "--parameterCategory", "0",
            "--parameterNumber", "0",
            "--level", "2",
            "-i", self.testdata + "/fc2020111303+0003grib2",
            "-o", "unittest_output_plot_grib2.png",
            "--debug"
            ]
        kwargs = surfex.parse_args_plot_field(argv)
        surfex.run_plot_field(**kwargs)

    def test_plot_netcdf(self):
        argv = [
            "-it", "netcdf",
            "-t", "2020111306",
            "-g", "test/settings/conf_proj_test.json",
            "-v", "air_temperature_2m",
            "-i", self.testdata + "/meps_det_2_5km_20201113T03Z.nc",
            "-o", "unittest_output_plot_nc.png",
            "--debug"
            ]
        kwargs = surfex.parse_args_plot_field(argv)
        surfex.run_plot_field(**kwargs)

    def test_plot_obs_frost_json(self):
        argv = [
            "-it", "obs",
            "--obs_type", "json",
            "-t", "2020111306",
            "-v", "air_temperature",
            "-i", self.testdata + "/unittest_frost_t2m.json",
            "-o", "unittest_output_plot_obs_frost_json.png",
            "--debug"
            ]
        kwargs = surfex.parse_args_plot_field(argv)
        surfex.run_plot_field(**kwargs)



import sys
from datetime import datetime


def error(message):
    """ Write error message to console and abort """
    print("\033[1;31mError: " + message + "\033[0m")
    sys.exit(1)


def info(message, level=0):
    """ Write a information message to console """
    if level < 1:
        print("\033[1;92mINFO: " + message + "\033[0m")


def warning(message):
    """ Write a warning message to console """
    print("\033[1;33mWarning: " + message + "\033[0m")


def unixtime_to_datenum(time):

    """ Converts unixtime into datenum

    Arguments:
      time (int): unixtime in seconds since 1970

    Returns:
      int: datenum value
    """
   
    dt = datetime.utcfromtimestamp(time)
    try:
        from matplotlib import dates
        dt2 = dates.date2num(dt)
        return dt2
    except ImportError:
        raise Exception("You need to have dates installed")


class YamlReaderError(Exception):
    pass


def data_merge(a, b):
    """merges b into a and return merged result

    NOTE: tuples and arbitrary objects are not handled as it is totally ambiguous what should happen"""
    key = None
    # ## debug output
    # sys.stderr.write("DEBUG: %s to %s\n" %(b,a))
    try:
        if a is None or isinstance(a, str) or isinstance(a, int) or isinstance(a, float):
            # border case for first run or if a is a primitive
            a = b
        elif isinstance(a, list):
            # lists can be only appended
            if isinstance(b, list):
                # merge lists
                a.extend(b)
            else:
                # append to list
                a.append(b)
        elif isinstance(a, dict):
            # dicts must be merged
            if isinstance(b, dict):
                for key in b:
                    if key in a:
                        a[key] = data_merge(a[key], b[key])
                    else:
                        a[key] = b[key]
            else:
                raise YamlReaderError('Cannot merge non-dict "%s" into dict "%s"' % (b, a))
        else:
            raise YamlReaderError('NOT IMPLEMENTED "%s" into "%s"' % (b, a))
    except TypeError as e:
        raise YamlReaderError('TypeError "%s" in key "%s" when merging "%s" into "%s"' % (e, key, b, a))
    return a



import unittest
import surfex


class FirstGuess4OiTest(unittest.TestCase):

    def setUp(self):
        self.testdata = "testdata/"

    def test_firstguess4oi_grib1(self):

        argv = [
            "-c", "surfex/cfg/first_guess.yml",
            "-i", self.testdata + "/fc2020111303+0003grib1",
            "-if", "grib1",
            "-dtg", "2020111306",
            "-d", "test/settings/conf_proj_test.json",
            "--laf_converter", "none",
            "--debug",
            "-o", "unittest_FirstGuess4gridpp_grib1.nc",
            "air_temperature_2m",
            "relative_humidity_2m",
            "surface_snow_thickness"
        ]
        kwargs = surfex.parse_args_first_guess_for_oi(argv)
        surfex.first_guess_for_oi(**kwargs)

    def test_firstguess4oi_grib2(self):

        argv = [
            "-c", "surfex/cfg/first_guess.yml",
            "-i", self.testdata + "/fc2020111303+0003grib2",
            "-if", "grib2",
            "-dtg", "2020111306",
            "-d", "test/settings/conf_proj_test.json",
            "--laf_converter", "none",
            "--debug",
            "-o", "unittest_FirstGuess4gridpp_grib2.nc",
            "air_temperature_2m",
            "relative_humidity_2m",
            "surface_snow_thickness"
        ]
        kwargs = surfex.parse_args_first_guess_for_oi(argv)
        surfex.first_guess_for_oi(**kwargs)
